EXPECTED = {'IEEE1609dot2': {'extensibility-implied': False,
                  'imports': {'IEEE1609dot2BaseTypes': ['CrlSeries',
                                                        'EccP256CurvePoint',
                                                        'EciesP256EncryptedKey',
                                                        'EncryptionKey',
                                                        'GeographicRegion',
                                                        'GroupLinkageValue',
                                                        'HashAlgorithm',
                                                        'HashedId3',
                                                        'HashedId8',
                                                        'HashedId32',
                                                        'Hostname',
                                                        'IValue',
                                                        'LinkageValue',
                                                        'Opaque',
                                                        'Psid',
                                                        'PsidSsp',
                                                        'PsidSspRange',
                                                        'PublicEncryptionKey',
                                                        'PublicVerificationKey',
                                                        'SequenceOfPsidSsp',
                                                        'SequenceOfPsidSspRange',
                                                        'ServiceSpecificPermissions',
                                                        'Signature',
                                                        'SubjectAssurance',
                                                        'SymmetricEncryptionKey',
                                                        'ThreeDLocation',
                                                        'Time64',
                                                        'Uint3',
                                                        'Uint8',
                                                        'Uint16',
                                                        'Uint32',
                                                        'ValidityPeriod']},
                  'object-classes': {},
                  'object-sets': {},
                  'tags': 'AUTOMATIC',
                  'types': {'AesCcmCiphertext': {'members': [{'name': 'nonce',
                                                              'size': [12],
                                                              'type': 'OCTET '
                                                                      'STRING'},
                                                             {'name': 'ccmCiphertext',
                                                              'type': 'Opaque'}],
                                                 'type': 'SEQUENCE'},
                            'Certificate': {'restricted-to': ['ImplicitCertificate',
                                                              'ExplicitCertificate'],
                                            'type': 'CertificateBase'},
                            'CertificateBase': {'members': [{'name': 'version',
                                                             'restricted-to': [3],
                                                             'type': 'Uint8'},
                                                            {'name': 'type',
                                                             'type': 'CertificateType'},
                                                            {'name': 'issuer',
                                                             'type': 'IssuerIdentifier'},
                                                            {'name': 'toBeSigned',
                                                             'type': 'ToBeSignedCertificate'},
                                                            {'name': 'signature',
                                                             'optional': True,
                                                             'type': 'Signature'}],
                                                'type': 'SEQUENCE'},
                            'CertificateId': {'members': [{'name': 'linkageData',
                                                           'type': 'LinkageData'},
                                                          {'name': 'name',
                                                           'type': 'Hostname'},
                                                          {'name': 'binaryId',
                                                           'size': [(1, 64)],
                                                           'type': 'OCTET '
                                                                   'STRING'},
                                                          {'name': 'none',
                                                           'type': 'NULL'},
                                                          None],
                                              'type': 'CHOICE'},
                            'CertificateType': {'type': 'ENUMERATED',
                                                'values': [('explicit', 0),
                                                           ('implicit', 1),
                                                           None]},
                            'Countersignature': {'type': 'Ieee1609Dot2Data',
                                                 'with-components': [None,
                                                                     ('content',
                                                                      {'with-components': [None,
                                                                                           ('signedData',
                                                                                            {'with-components': [None,
                                                                                                                 ('tbsData',
                                                                                                                  {'with-components': [None,
                                                                                                                                       ('payload',
                                                                                                                                        {'with-components': [None,
                                                                                                                                                             ('data',
                                                                                                                                                              'ABSENT'),
                                                                                                                                                             ('extDataHash',
                                                                                                                                                              'PRESENT')]}),
                                                                                                                                       ('headerInfo',
                                                                                                                                        {'with-components': [None,
                                                                                                                                                             ('generationTime',
                                                                                                                                                              'PRESENT'),
                                                                                                                                                             ('expiryTime',
                                                                                                                                                              'ABSENT'),
                                                                                                                                                             ('generationLocation',
                                                                                                                                                              'ABSENT'),
                                                                                                                                                             ('p2pcdLearningRequest',
                                                                                                                                                              'ABSENT'),
                                                                                                                                                             ('missingCrlIdentifier',
                                                                                                                                                              'ABSENT'),
                                                                                                                                                             ('encryptionKey',
                                                                                                                                                              'ABSENT')]})]})]})]})]},
                            'EncryptedData': {'members': [{'name': 'recipients',
                                                           'type': 'SequenceOfRecipientInfo'},
                                                          {'name': 'ciphertext',
                                                           'type': 'SymmetricCiphertext'}],
                                              'type': 'SEQUENCE'},
                            'EncryptedDataEncryptionKey': {'members': [{'name': 'eciesNistP256',
                                                                        'type': 'EciesP256EncryptedKey'},
                                                                       {'name': 'eciesBrainpoolP256r1',
                                                                        'type': 'EciesP256EncryptedKey'},
                                                                       None],
                                                           'type': 'CHOICE'},
                            'EndEntityType': {'named-bits': [('app', '0'),
                                                             ('enrol', '1')],
                                              'size': [8],
                                              'type': 'BIT STRING'},
                            'ExplicitCertificate': {'type': 'CertificateBase',
                                                    'with-components': [None,
                                                                        ('type',
                                                                         'explicit'),
                                                                        ('toBeSigned',
                                                                         {'with-components': [None,
                                                                                              ('verifyKeyIndicator',
                                                                                               {'with-components': [['verificationKey']]})]}),
                                                                        ('signature',
                                                                         'PRESENT')]},
                            'HashedData': {'members': [{'name': 'sha256HashedData',
                                                        'size': [32],
                                                        'type': 'OCTET STRING'},
                                                       None],
                                           'type': 'CHOICE'},
                            'HeaderInfo': {'members': [{'name': 'psid',
                                                        'type': 'Psid'},
                                                       {'name': 'generationTime',
                                                        'optional': True,
                                                        'type': 'Time64'},
                                                       {'name': 'expiryTime',
                                                        'optional': True,
                                                        'type': 'Time64'},
                                                       {'name': 'generationLocation',
                                                        'optional': True,
                                                        'type': 'ThreeDLocation'},
                                                       {'name': 'p2pcdLearningRequest',
                                                        'optional': True,
                                                        'type': 'HashedId3'},
                                                       {'name': 'missingCrlIdentifier',
                                                        'optional': True,
                                                        'type': 'MissingCrlIdentifier'},
                                                       {'name': 'encryptionKey',
                                                        'optional': True,
                                                        'type': 'EncryptionKey'},
                                                       None],
                                           'type': 'SEQUENCE'},
                            'Ieee1609Dot2Content': {'members': [{'name': 'unsecuredData',
                                                                 'type': 'Opaque'},
                                                                {'name': 'signedData',
                                                                 'type': 'SignedData'},
                                                                {'name': 'encryptedData',
                                                                 'type': 'EncryptedData'},
                                                                {'name': 'signedCertificateRequest',
                                                                 'type': 'Opaque'},
                                                                None],
                                                    'type': 'CHOICE'},
                            'Ieee1609Dot2Data': {'members': [{'name': 'protocolVersion',
                                                              'restricted-to': [3],
                                                              'type': 'Uint8'},
                                                             {'name': 'content',
                                                              'type': 'Ieee1609Dot2Content'}],
                                                 'type': 'SEQUENCE'},
                            'ImplicitCertificate': {'type': 'CertificateBase',
                                                    'with-components': [None,
                                                                        ('type',
                                                                         'implicit'),
                                                                        ('toBeSigned',
                                                                         {'with-components': [None,
                                                                                              ('verifyKeyIndicator',
                                                                                               {'with-components': [['reconstructionValue']]})]}),
                                                                        ('signature',
                                                                         'ABSENT')]},
                            'IssuerIdentifier': {'members': [{'name': 'sha256AndDigest',
                                                              'type': 'HashedId8'},
                                                             {'name': 'self',
                                                              'type': 'HashAlgorithm'},
                                                             None],
                                                 'type': 'CHOICE'},
                            'LinkageData': {'members': [{'name': 'iCert',
                                                         'type': 'IValue'},
                                                        {'name': 'linkage-value',
                                                         'type': 'LinkageValue'},
                                                        {'name': 'group-linkage-value',
                                                         'optional': True,
                                                         'type': 'GroupLinkageValue'}],
                                            'type': 'SEQUENCE'},
                            'MissingCrlIdentifier': {'members': [{'name': 'cracaId',
                                                                  'type': 'HashedId3'},
                                                                 {'name': 'crlSeries',
                                                                  'type': 'CrlSeries'},
                                                                 None],
                                                     'type': 'SEQUENCE'},
                            'PKRecipientInfo': {'members': [{'name': 'recipientId',
                                                             'type': 'HashedId8'},
                                                            {'name': 'encKey',
                                                             'type': 'EncryptedDataEncryptionKey'}],
                                                'type': 'SEQUENCE'},
                            'PreSharedKeyRecipientInfo': {'type': 'HashedId8'},
                            'PsidGroupPermissions': {'members': [{'name': 'appPermissions',
                                                                  'type': 'SubjectPermissions'},
                                                                 {'default': 1,
                                                                  'name': 'minChainDepth',
                                                                  'type': 'INTEGER'},
                                                                 {'default': 0,
                                                                  'name': 'chainDepthRange',
                                                                  'type': 'INTEGER'},
                                                                 {'default': ['app'],
                                                                  'name': 'eeType',
                                                                  'type': 'EndEntityType'}],
                                                     'type': 'SEQUENCE'},
                            'RecipientInfo': {'members': [{'name': 'pskRecipInfo',
                                                           'type': 'PreSharedKeyRecipientInfo'},
                                                          {'name': 'symmRecipInfo',
                                                           'type': 'SymmRecipientInfo'},
                                                          {'name': 'certRecipInfo',
                                                           'type': 'PKRecipientInfo'},
                                                          {'name': 'signedDataRecipInfo',
                                                           'type': 'PKRecipientInfo'},
                                                          {'name': 'rekRecipInfo',
                                                           'type': 'PKRecipientInfo'}],
                                              'type': 'CHOICE'},
                            'SequenceOfCertificate': {'element': {'type': 'Certificate'},
                                                      'type': 'SEQUENCE OF'},
                            'SequenceOfPsidGroupPermissions': {'element': {'type': 'PsidGroupPermissions'},
                                                               'type': 'SEQUENCE '
                                                                       'OF'},
                            'SequenceOfRecipientInfo': {'element': {'type': 'RecipientInfo'},
                                                        'type': 'SEQUENCE OF'},
                            'SignedData': {'members': [{'name': 'hashId',
                                                        'type': 'HashAlgorithm'},
                                                       {'name': 'tbsData',
                                                        'type': 'ToBeSignedData'},
                                                       {'name': 'signer',
                                                        'type': 'SignerIdentifier'},
                                                       {'name': 'signature',
                                                        'type': 'Signature'}],
                                           'type': 'SEQUENCE'},
                            'SignedDataPayload': {'members': [{'name': 'data',
                                                               'optional': True,
                                                               'type': 'Ieee1609Dot2Data'},
                                                              {'name': 'extDataHash',
                                                               'optional': True,
                                                               'type': 'HashedData'},
                                                              None],
                                                  'type': 'SEQUENCE',
                                                  'with-components': [None,
                                                                      ('extDataHash',
                                                                       'PRESENT')]},
                            'SignerIdentifier': {'members': [{'name': 'digest',
                                                              'type': 'HashedId8'},
                                                             {'name': 'certificate',
                                                              'type': 'SequenceOfCertificate'},
                                                             {'name': 'self',
                                                              'type': 'NULL'},
                                                             None],
                                                 'type': 'CHOICE'},
                            'SubjectPermissions': {'members': [{'name': 'explicit',
                                                                'type': 'SequenceOfPsidSspRange'},
                                                               {'name': 'all',
                                                                'type': 'NULL'},
                                                               None],
                                                   'type': 'CHOICE'},
                            'SymmRecipientInfo': {'members': [{'name': 'recipientId',
                                                               'type': 'HashedId8'},
                                                              {'name': 'encKey',
                                                               'type': 'SymmetricCiphertext'}],
                                                  'type': 'SEQUENCE'},
                            'SymmetricCiphertext': {'members': [{'name': 'aes128ccm',
                                                                 'type': 'AesCcmCiphertext'},
                                                                None],
                                                    'type': 'CHOICE'},
                            'ToBeSignedCertificate': {'members': [{'name': 'id',
                                                                   'type': 'CertificateId'},
                                                                  {'name': 'cracaId',
                                                                   'type': 'HashedId3'},
                                                                  {'name': 'crlSeries',
                                                                   'type': 'CrlSeries'},
                                                                  {'name': 'validityPeriod',
                                                                   'type': 'ValidityPeriod'},
                                                                  {'name': 'region',
                                                                   'optional': True,
                                                                   'type': 'GeographicRegion'},
                                                                  {'name': 'assuranceLevel',
                                                                   'optional': True,
                                                                   'type': 'SubjectAssurance'},
                                                                  {'name': 'appPermissions',
                                                                   'optional': True,
                                                                   'type': 'SequenceOfPsidSsp'},
                                                                  {'name': 'certIssuePermissions',
                                                                   'optional': True,
                                                                   'type': 'SequenceOfPsidGroupPermissions'},
                                                                  {'name': 'certRequestPermissions',
                                                                   'optional': True,
                                                                   'type': 'SequenceOfPsidGroupPermissions'},
                                                                  {'name': 'canRequestRollover',
                                                                   'optional': True,
                                                                   'type': 'NULL'},
                                                                  {'name': 'encryptionKey',
                                                                   'optional': True,
                                                                   'type': 'PublicEncryptionKey'},
                                                                  {'name': 'verifyKeyIndicator',
                                                                   'type': 'VerificationKeyIndicator'},
                                                                  None],
                                                      'type': 'SEQUENCE',
                                                      'with-components': [None,
                                                                          ('certRequestPermissions',
                                                                           'PRESENT')]},
                            'ToBeSignedData': {'members': [{'name': 'payload',
                                                            'type': 'SignedDataPayload'},
                                                           {'name': 'headerInfo',
                                                            'type': 'HeaderInfo'}],
                                               'type': 'SEQUENCE'},
                            'VerificationKeyIndicator': {'members': [{'name': 'verificationKey',
                                                                      'type': 'PublicVerificationKey'},
                                                                     {'name': 'reconstructionValue',
                                                                      'type': 'EccP256CurvePoint'},
                                                                     None],
                                                         'type': 'CHOICE'}},
                  'values': {}},
 'IEEE1609dot2-Peer2Peer': {'extensibility-implied': False,
                            'imports': {'IEEE1609dot2': ['Certificate'],
                                        'IEEE1609dot2BaseTypes': ['Uint8']},
                            'object-classes': {},
                            'object-sets': {},
                            'tags': 'AUTOMATIC',
                            'types': {'CaCertP2pPDU': {'element': {'type': 'Certificate'},
                                                       'type': 'SEQUENCE OF'},
                                      'Ieee1609dot2Peer2PeerPDU': {'members': [{'name': 'version',
                                                                                'restricted-to': [1],
                                                                                'type': 'Uint8'},
                                                                               {'members': [{'name': 'caCerts',
                                                                                             'type': 'CaCertP2pPDU'},
                                                                                            None],
                                                                                'name': 'content',
                                                                                'type': 'CHOICE'}],
                                                                   'type': 'SEQUENCE'}},
                            'values': {}},
 'IEEE1609dot2BaseTypes': {'extensibility-implied': False,
                           'imports': {},
                           'object-classes': {},
                           'object-sets': {},
                           'tags': 'AUTOMATIC',
                           'types': {'BasePublicEncryptionKey': {'members': [{'name': 'eciesNistP256',
                                                                              'type': 'EccP256CurvePoint'},
                                                                             {'name': 'eciesBrainpoolP256r1',
                                                                              'type': 'EccP256CurvePoint'},
                                                                             None],
                                                                 'type': 'CHOICE'},
                                     'CircularRegion': {'members': [{'name': 'center',
                                                                     'type': 'TwoDLocation'},
                                                                    {'name': 'radius',
                                                                     'type': 'Uint16'}],
                                                        'type': 'SEQUENCE'},
                                     'CountryAndRegions': {'members': [{'name': 'countryOnly',
                                                                        'type': 'CountryOnly'},
                                                                       {'name': 'regions',
                                                                        'type': 'SequenceOfUint8'}],
                                                           'type': 'SEQUENCE'},
                                     'CountryAndSubregions': {'members': [{'name': 'country',
                                                                           'type': 'CountryOnly'},
                                                                          {'name': 'regionAndSubregions',
                                                                           'type': 'SequenceOfRegionAndSubregions'}],
                                                              'type': 'SEQUENCE'},
                                     'CountryOnly': {'type': 'Uint16'},
                                     'CrlSeries': {'type': 'Uint16'},
                                     'Duration': {'members': [{'name': 'microseconds',
                                                               'type': 'Uint16'},
                                                              {'name': 'milliseconds',
                                                               'type': 'Uint16'},
                                                              {'name': 'seconds',
                                                               'type': 'Uint16'},
                                                              {'name': 'minutes',
                                                               'type': 'Uint16'},
                                                              {'name': 'hours',
                                                               'type': 'Uint16'},
                                                              {'name': 'sixtyHours',
                                                               'type': 'Uint16'},
                                                              {'name': 'years',
                                                               'type': 'Uint16'}],
                                                  'type': 'CHOICE'},
                                     'EccP256CurvePoint': {'members': [{'name': 'x-only',
                                                                        'size': [32],
                                                                        'type': 'OCTET '
                                                                                'STRING'},
                                                                       {'name': 'fill',
                                                                        'type': 'NULL'},
                                                                       {'name': 'compressed-y-0',
                                                                        'size': [32],
                                                                        'type': 'OCTET '
                                                                                'STRING'},
                                                                       {'name': 'compressed-y-1',
                                                                        'size': [32],
                                                                        'type': 'OCTET '
                                                                                'STRING'},
                                                                       {'members': [{'name': 'x',
                                                                                     'size': [32],
                                                                                     'type': 'OCTET '
                                                                                             'STRING'},
                                                                                    {'name': 'y',
                                                                                     'size': [32],
                                                                                     'type': 'OCTET '
                                                                                             'STRING'}],
                                                                        'name': 'uncompressed',
                                                                        'type': 'SEQUENCE'}],
                                                           'type': 'CHOICE'},
                                     'EcdsaP256Signature': {'members': [{'name': 'r',
                                                                         'type': 'EccP256CurvePoint'},
                                                                        {'name': 's',
                                                                         'size': [32],
                                                                         'type': 'OCTET '
                                                                                 'STRING'}],
                                                            'type': 'SEQUENCE'},
                                     'EciesP256EncryptedKey': {'members': [{'name': 'v',
                                                                            'type': 'EccP256CurvePoint'},
                                                                           {'name': 'c',
                                                                            'size': [16],
                                                                            'type': 'OCTET '
                                                                                    'STRING'},
                                                                           {'name': 't',
                                                                            'size': [16],
                                                                            'type': 'OCTET '
                                                                                    'STRING'}],
                                                               'type': 'SEQUENCE'},
                                     'ElevInt': {'type': 'Uint16'},
                                     'Elevation': {'type': 'ElevInt'},
                                     'EncryptionKey': {'members': [{'name': 'public',
                                                                    'type': 'PublicEncryptionKey'},
                                                                   {'name': 'symmetric',
                                                                    'type': 'SymmetricEncryptionKey'}],
                                                       'type': 'CHOICE'},
                                     'GeographicRegion': {'members': [{'name': 'circularRegion',
                                                                       'type': 'CircularRegion'},
                                                                      {'name': 'rectangularRegion',
                                                                       'type': 'SequenceOfRectangularRegion'},
                                                                      {'name': 'polygonalRegion',
                                                                       'type': 'PolygonalRegion'},
                                                                      {'name': 'identifiedRegion',
                                                                       'type': 'SequenceOfIdentifiedRegion'},
                                                                      None],
                                                          'type': 'CHOICE'},
                                     'GroupLinkageValue': {'members': [{'name': 'jValue',
                                                                        'size': [4],
                                                                        'type': 'OCTET '
                                                                                'STRING'},
                                                                       {'name': 'value',
                                                                        'size': [9],
                                                                        'type': 'OCTET '
                                                                                'STRING'}],
                                                           'type': 'SEQUENCE'},
                                     'HashAlgorithm': {'type': 'ENUMERATED',
                                                       'values': [('sha256', 0),
                                                                  None]},
                                     'HashedId10': {'size': [10],
                                                    'type': 'OCTET STRING'},
                                     'HashedId3': {'size': [3],
                                                   'type': 'OCTET STRING'},
                                     'HashedId32': {'size': [32],
                                                    'type': 'OCTET STRING'},
                                     'HashedId4': {'size': [4],
                                                   'type': 'OCTET STRING'},
                                     'HashedId8': {'size': [8],
                                                   'type': 'OCTET STRING'},
                                     'Hostname': {'size': [(0, 255)],
                                                  'type': 'UTF8String'},
                                     'IValue': {'type': 'Uint16'},
                                     'IdentifiedRegion': {'members': [{'name': 'countryOnly',
                                                                       'type': 'CountryOnly'},
                                                                      {'name': 'countryAndRegions',
                                                                       'type': 'CountryAndRegions'},
                                                                      {'name': 'countryAndSubregions',
                                                                       'type': 'CountryAndSubregions'},
                                                                      None],
                                                          'type': 'CHOICE'},
                                     'KnownLatitude': {'restricted-to': [(-900000000,
                                                                          900000000)],
                                                       'type': 'NinetyDegreeInt'},
                                     'KnownLongitude': {'restricted-to': [(-1799999999,
                                                                           1800000000)],
                                                        'type': 'OneEightyDegreeInt'},
                                     'LaId': {'size': [2],
                                              'type': 'OCTET STRING'},
                                     'Latitude': {'type': 'NinetyDegreeInt'},
                                     'LinkageSeed': {'size': [16],
                                                     'type': 'OCTET STRING'},
                                     'LinkageValue': {'size': [9],
                                                      'type': 'OCTET STRING'},
                                     'Longitude': {'type': 'OneEightyDegreeInt'},
                                     'NinetyDegreeInt': {'restricted-to': [(-900000000,
                                                                            900000001)],
                                                         'values': [('min', -900000000),
                                                                    ('max', 900000000),
                                                                    ('unknown', 900000001)],
                                                         'type': 'INTEGER'},
                                     'OneEightyDegreeInt': {'restricted-to': [(-1799999999,
                                                                               1800000001)],
                                                            'values': [('min', -1799999999),
                                                                       ('max', 1800000000),
                                                                       ('unknown', 1800000001)],
                                                            'type': 'INTEGER'},
                                     'Opaque': {'type': 'OCTET STRING'},
                                     'PolygonalRegion': {'element': {'type': 'TwoDLocation'},
                                                         'size': [(3, 'MAX')],
                                                         'type': 'SEQUENCE OF'},
                                     'Psid': {'restricted-to': [(0, 'MAX')],
                                              'type': 'INTEGER'},
                                     'PsidSsp': {'members': [{'name': 'psid',
                                                              'type': 'Psid'},
                                                             {'name': 'ssp',
                                                              'optional': True,
                                                              'type': 'ServiceSpecificPermissions'}],
                                                 'type': 'SEQUENCE'},
                                     'PsidSspRange': {'members': [{'name': 'psid',
                                                                   'type': 'Psid'},
                                                                  {'name': 'sspRange',
                                                                   'optional': True,
                                                                   'type': 'SspRange'}],
                                                      'type': 'SEQUENCE'},
                                     'PublicEncryptionKey': {'members': [{'name': 'supportedSymmAlg',
                                                                          'type': 'SymmAlgorithm'},
                                                                         {'name': 'publicKey',
                                                                          'type': 'BasePublicEncryptionKey'}],
                                                             'type': 'SEQUENCE'},
                                     'PublicVerificationKey': {'members': [{'name': 'ecdsaNistP256',
                                                                            'type': 'EccP256CurvePoint'},
                                                                           {'name': 'ecdsaBrainpoolP256r1',
                                                                            'type': 'EccP256CurvePoint'},
                                                                           None],
                                                               'type': 'CHOICE'},
                                     'RectangularRegion': {'members': [{'name': 'northWest',
                                                                        'type': 'TwoDLocation'},
                                                                       {'name': 'southEast',
                                                                        'type': 'TwoDLocation'}],
                                                           'type': 'SEQUENCE'},
                                     'RegionAndSubregions': {'members': [{'name': 'region',
                                                                          'type': 'Uint8'},
                                                                         {'name': 'subregions',
                                                                          'type': 'SequenceOfUint16'}],
                                                             'type': 'SEQUENCE'},
                                     'SequenceOfIdentifiedRegion': {'element': {'type': 'IdentifiedRegion'},
                                                                    'type': 'SEQUENCE '
                                                                            'OF'},
                                     'SequenceOfOctetString': {'element': {'size': [(0,
                                                                                     'MAX')],
                                                                           'type': 'OCTET '
                                                                                   'STRING'},
                                                               'size': [(0,
                                                                         'MAX')],
                                                               'type': 'SEQUENCE '
                                                                       'OF'},
                                     'SequenceOfPsid': {'element': {'type': 'Psid'},
                                                        'type': 'SEQUENCE OF'},
                                     'SequenceOfPsidSsp': {'element': {'type': 'PsidSsp'},
                                                           'type': 'SEQUENCE '
                                                                   'OF'},
                                     'SequenceOfPsidSspRange': {'element': {'type': 'PsidSspRange'},
                                                                'type': 'SEQUENCE '
                                                                        'OF'},
                                     'SequenceOfRectangularRegion': {'element': {'type': 'RectangularRegion'},
                                                                     'type': 'SEQUENCE '
                                                                             'OF'},
                                     'SequenceOfRegionAndSubregions': {'element': {'type': 'RegionAndSubregions'},
                                                                       'type': 'SEQUENCE '
                                                                               'OF'},
                                     'SequenceOfUint16': {'element': {'type': 'Uint16'},
                                                          'type': 'SEQUENCE '
                                                                  'OF'},
                                     'SequenceOfUint3': {'element': {'type': 'Uint3'},
                                                         'type': 'SEQUENCE OF'},
                                     'SequenceOfUint8': {'element': {'type': 'Uint8'},
                                                         'type': 'SEQUENCE OF'},
                                     'ServiceSpecificPermissions': {'members': [{'name': 'opaque',
                                                                                 'size': [(0,
                                                                                           'MAX')],
                                                                                 'type': 'OCTET '
                                                                                         'STRING'},
                                                                                None],
                                                                    'type': 'CHOICE'},
                                     'Signature': {'members': [{'name': 'ecdsaNistP256Signature',
                                                                'type': 'EcdsaP256Signature'},
                                                               {'name': 'ecdsaBrainpoolP256r1Signature',
                                                                'type': 'EcdsaP256Signature'},
                                                               None],
                                                   'type': 'CHOICE'},
                                     'SspRange': {'members': [{'name': 'opaque',
                                                               'type': 'SequenceOfOctetString'},
                                                              {'name': 'all',
                                                               'type': 'NULL'},
                                                              None],
                                                  'type': 'CHOICE'},
                                     'SubjectAssurance': {'size': [1],
                                                          'type': 'OCTET '
                                                                  'STRING'},
                                     'SymmAlgorithm': {'type': 'ENUMERATED',
                                                       'values': [('aes128Ccm',
                                                                   0),
                                                                  None]},
                                     'SymmetricEncryptionKey': {'members': [{'name': 'aes128Ccm',
                                                                             'size': [16],
                                                                             'type': 'OCTET '
                                                                                     'STRING'},
                                                                            None],
                                                                'type': 'CHOICE'},
                                     'ThreeDLocation': {'members': [{'name': 'latitude',
                                                                     'type': 'Latitude'},
                                                                    {'name': 'longitude',
                                                                     'type': 'Longitude'},
                                                                    {'name': 'elevation',
                                                                     'type': 'Elevation'}],
                                                        'type': 'SEQUENCE'},
                                     'Time32': {'type': 'Uint32'},
                                     'Time64': {'type': 'Uint64'},
                                     'TwoDLocation': {'members': [{'name': 'latitude',
                                                                   'type': 'Latitude'},
                                                                  {'name': 'longitude',
                                                                   'type': 'Longitude'}],
                                                      'type': 'SEQUENCE'},
                                     'Uint16': {'restricted-to': [(0, 65535)],
                                                'type': 'INTEGER'},
                                     'Uint3': {'restricted-to': [(0, 7)],
                                               'type': 'INTEGER'},
                                     'Uint32': {'restricted-to': [(0,
                                                                   4294967295)],
                                                'type': 'INTEGER'},
                                     'Uint64': {'restricted-to': [(0,
                                                                   18446744073709551615)],
                                                'type': 'INTEGER'},
                                     'Uint8': {'restricted-to': [(0, 255)],
                                               'type': 'INTEGER'},
                                     'UnknownLatitude': {'restricted-to': [900000001],
                                                         'type': 'NinetyDegreeInt'},
                                     'UnknownLongitude': {'restricted-to': [1800000001],
                                                          'type': 'OneEightyDegreeInt'},
                                     'ValidityPeriod': {'members': [{'name': 'start',
                                                                     'type': 'Time32'},
                                                                    {'name': 'duration',
                                                                     'type': 'Duration'}],
                                                        'type': 'SEQUENCE'}},
                           'values': {}},
 'IEEE1609dot2Crl': {'extensibility-implied': False,
                     'imports': {'IEEE1609dot2': ['Ieee1609Dot2Data'],
                                 'IEEE1609dot2BaseTypes': ['Psid', 'Opaque'],
                                 'IEEE1609dot2CrlBaseTypes': ['CrlContents']},
                     'object-classes': {},
                     'object-sets': {},
                     'tags': 'AUTOMATIC',
                     'types': {'CrlPsid': {'restricted-to': [135],
                                           'type': 'Psid'},
                               'SecuredCrl': {'type': 'Ieee1609Dot2Data',
                                              'with-components': [None,
                                                                  ('content',
                                                                   {'with-components': [('signedData',
                                                                                         {'with-components': [None,
                                                                                                              ('tbsData',
                                                                                                               {'with-components': [('payload',
                                                                                                                                     {'with-components': [None,
                                                                                                                                                          ('data',
                                                                                                                                                           {'with-components': [None,
                                                                                                                                                                                ('content',
                                                                                                                                                                                 {'with-components': [['unsecuredData',
                                                                                                                                                                                                       'CONTAINING',
                                                                                                                                                                                                       [{'type': 'CrlContents'},
                                                                                                                                                                                                        []]]]})]})]}),
                                                                                                                                    ('headerInfo',
                                                                                                                                     {'with-components': [None,
                                                                                                                                                          ('psid',
                                                                                                                                                           'CrlPsid'),
                                                                                                                                                          ('generationTime',
                                                                                                                                                           'ABSENT'),
                                                                                                                                                          ('expiryTime',
                                                                                                                                                           'ABSENT'),
                                                                                                                                                          ('generationLocation',
                                                                                                                                                           'ABSENT'),
                                                                                                                                                          ('p2pcdLearningRequest',
                                                                                                                                                           'ABSENT'),
                                                                                                                                                          ('missingCrlIdentifier',
                                                                                                                                                           'ABSENT'),
                                                                                                                                                          ('encryptionKey',
                                                                                                                                                           'ABSENT')]})]})]})]})]}},
                     'values': {}},
 'IEEE1609dot2CrlBaseTypes': {'extensibility-implied': False,
                              'imports': {'IEEE1609dot2BaseTypes': ['CrlSeries',
                                                                    'GeographicRegion',
                                                                    'HashedId8',
                                                                    'HashedId10',
                                                                    'IValue',
                                                                    'LaId',
                                                                    'LinkageSeed',
                                                                    'Opaque',
                                                                    'Psid',
                                                                    'Signature',
                                                                    'Time32',
                                                                    'Uint3',
                                                                    'Uint8',
                                                                    'Uint16',
                                                                    'Uint32',
                                                                    'ValidityPeriod']},
                              'object-classes': {},
                              'object-sets': {},
                              'tags': 'AUTOMATIC',
                              'types': {'CrlContents': {'members': [{'name': 'version',
                                                                     'restricted-to': [1],
                                                                     'type': 'Uint8'},
                                                                    {'name': 'crlSeries',
                                                                     'type': 'CrlSeries'},
                                                                    {'name': 'cracaId',
                                                                     'type': 'HashedId8'},
                                                                    {'name': 'issueDate',
                                                                     'type': 'Time32'},
                                                                    {'name': 'nextCrl',
                                                                     'type': 'Time32'},
                                                                    {'name': 'priorityInfo',
                                                                     'type': 'CrlPriorityInfo'},
                                                                    {'members': [{'name': 'fullHashCrl',
                                                                                  'type': 'ToBeSignedHashIdCrl'},
                                                                                 {'name': 'deltaHashCrl',
                                                                                  'type': 'ToBeSignedHashIdCrl'},
                                                                                 {'name': 'fullLinkedCrl',
                                                                                  'type': 'ToBeSignedLinkageValueCrl'},
                                                                                 {'name': 'deltaLinkedCrl',
                                                                                  'type': 'ToBeSignedLinkageValueCrl'},
                                                                                 None],
                                                                     'name': 'typeSpecific',
                                                                     'type': 'CHOICE'}],
                                                        'type': 'SEQUENCE'},
                                        'CrlPriorityInfo': {'members': [{'name': 'priority',
                                                                         'optional': True,
                                                                         'type': 'Uint8'},
                                                                        None],
                                                            'type': 'SEQUENCE'},
                                        'GroupCrlEntry': {'members': [{'name': 'iMax',
                                                                       'type': 'Uint16'},
                                                                      {'name': 'la1Id',
                                                                       'type': 'LaId'},
                                                                      {'name': 'linkageSeed1',
                                                                       'type': 'LinkageSeed'},
                                                                      {'name': 'la2Id',
                                                                       'type': 'LaId'},
                                                                      {'name': 'linkageSeed2',
                                                                       'type': 'LinkageSeed'}],
                                                          'type': 'SEQUENCE'},
                                        'HashBasedRevocationInfo': {'members': [{'name': 'id',
                                                                                 'type': 'HashedId10'},
                                                                                {'name': 'expiry',
                                                                                 'type': 'Time32'}],
                                                                    'type': 'SEQUENCE'},
                                        'IMaxGroup': {'members': [{'name': 'iMax',
                                                                   'type': 'Uint16'},
                                                                  {'name': 'contents',
                                                                   'type': 'SequenceOfIndividualRevocation'}],
                                                      'type': 'SEQUENCE'},
                                        'IndividualRevocation': {'members': [{'name': 'linkage-seed1',
                                                                              'type': 'LinkageSeed'},
                                                                             {'name': 'linkage-seed2',
                                                                              'type': 'LinkageSeed'}],
                                                                 'type': 'SEQUENCE'},
                                        'JMaxGroup': {'members': [{'name': 'jmax',
                                                                   'type': 'Uint8'},
                                                                  {'element': {'type': 'LAGroup'},
                                                                   'name': 'contents',
                                                                   'type': 'SEQUENCE '
                                                                           'OF'}],
                                                      'type': 'SEQUENCE'},
                                        'LAGroup': {'members': [{'name': 'la1Id',
                                                                 'type': 'LaId'},
                                                                {'name': 'la2Id',
                                                                 'type': 'LaId'},
                                                                {'name': 'contents',
                                                                 'type': 'SequenceOfIMaxGroup'}],
                                                    'type': 'SEQUENCE'},
                                        'SequenceOfGroupCrlEntry': {'element': {'type': 'GroupCrlEntry'},
                                                                    'type': 'SEQUENCE '
                                                                            'OF'},
                                        'SequenceOfHashBasedRevocationInfo': {'element': {'type': 'HashBasedRevocationInfo'},
                                                                              'type': 'SEQUENCE '
                                                                                      'OF'},
                                        'SequenceOfIMaxGroup': {'element': {'type': 'IMaxGroup'},
                                                                'type': 'SEQUENCE '
                                                                        'OF'},
                                        'SequenceOfIndividualRevocation': {'element': {'type': 'IndividualRevocation'},
                                                                           'type': 'SEQUENCE '
                                                                                   'OF'},
                                        'SequenceOfJMaxGroup': {'element': {'type': 'JMaxGroup'},
                                                                'type': 'SEQUENCE '
                                                                        'OF'},
                                        'SequenceOfLAGroup': {'element': {'type': 'LAGroup'},
                                                              'type': 'SEQUENCE '
                                                                      'OF'},
                                        'ToBeSignedHashIdCrl': {'members': [{'name': 'crlSerial',
                                                                             'type': 'Uint32'},
                                                                            {'name': 'entries',
                                                                             'type': 'SequenceOfHashBasedRevocationInfo'},
                                                                            None],
                                                                'type': 'SEQUENCE'},
                                        'ToBeSignedLinkageValueCrl': {'members': [{'name': 'iRev',
                                                                                   'type': 'IValue'},
                                                                                  {'name': 'indexWithinI',
                                                                                   'type': 'Uint8'},
                                                                                  {'name': 'individual',
                                                                                   'optional': True,
                                                                                   'type': 'SequenceOfJMaxGroup'},
                                                                                  {'name': 'groups',
                                                                                   'optional': True,
                                                                                   'type': 'SequenceOfGroupCrlEntry'},
                                                                                  None],
                                                                      'type': 'SEQUENCE',
                                                                      'with-components': [None,
                                                                                          ('groups',
                                                                                           'PRESENT')]}},
                              'values': {}},
 'IEEE1609dot2CrlSsp': {'extensibility-implied': False,
                        'imports': {'IEEE1609dot2BaseTypes': ['CrlSeries',
                                                              'Uint8']},
                        'object-classes': {},
                        'object-sets': {},
                        'tags': 'AUTOMATIC',
                        'types': {'CracaType': {'type': 'ENUMERATED',
                                                'values': [('isCraca', 0),
                                                           ('issuerIsCraca',
                                                            1)]},
                                  'CrlSsp': {'members': [{'name': 'version',
                                                          'restricted-to': [1],
                                                          'type': 'Uint8'},
                                                         {'name': 'associatedCraca',
                                                          'type': 'CracaType'},
                                                         {'name': 'crls',
                                                          'type': 'PermissibleCrls'},
                                                         None],
                                             'type': 'SEQUENCE'},
                                  'PermissibleCrls': {'element': {'type': 'CrlSeries'},
                                                      'type': 'SEQUENCE OF'}},
                        'values': {}}}



EXPECTED = {'PKIX1Explicit88': {'extensibility-implied': False,
                     'imports': {},
                     'object-classes': {},
                     'object-sets': {},
                     'tags': 'EXPLICIT',
                     'types': {'AdministrationDomainName': {'members': [{'name': 'numeric',
                                                                         'size': [(0,
                                                                                   'ub-domain-name-length')],
                                                                         'type': 'NumericString'},
                                                                        {'name': 'printable',
                                                                         'size': [(0,
                                                                                   'ub-domain-name-length')],
                                                                         'type': 'PrintableString'}],
                                                            'tag': {'class': 'APPLICATION',
                                                                    'number': 2},
                                                            'type': 'CHOICE'},
                               'AlgorithmIdentifier': {'members': [{'name': 'algorithm',
                                                                    'type': 'OBJECT '
                                                                            'IDENTIFIER'},
                                                                   {'choices': {},
                                                                    'name': 'parameters',
                                                                    'optional': True,
                                                                    'type': 'ANY '
                                                                            'DEFINED '
                                                                            'BY',
                                                                    'value': 'algorithm'}],
                                                       'type': 'SEQUENCE'},
                               'Attribute': {'members': [{'name': 'type',
                                                          'type': 'AttributeType'},
                                                         {'element': {'type': 'AttributeValue'},
                                                          'name': 'values',
                                                          'type': 'SET OF'}],
                                             'type': 'SEQUENCE'},
                               'AttributeType': {'type': 'OBJECT IDENTIFIER'},
                               'AttributeTypeAndValue': {'members': [{'name': 'type',
                                                                      'type': 'AttributeType'},
                                                                     {'name': 'value',
                                                                      'type': 'AttributeValue'}],
                                                         'type': 'SEQUENCE'},
                               'AttributeValue': {'type': 'ANY'},
                               'BuiltInDomainDefinedAttribute': {'members': [{'name': 'type',
                                                                              'size': [(1,
                                                                                        'ub-domain-defined-attribute-type-length')],
                                                                              'type': 'PrintableString'},
                                                                             {'name': 'value',
                                                                              'size': [(1,
                                                                                        'ub-domain-defined-attribute-value-length')],
                                                                              'type': 'PrintableString'}],
                                                                 'type': 'SEQUENCE'},
                               'BuiltInDomainDefinedAttributes': {'element': {'type': 'BuiltInDomainDefinedAttribute'},
                                                                  'size': [(1,
                                                                            'ub-domain-defined-attributes')],
                                                                  'type': 'SEQUENCE '
                                                                          'OF'},
                               'BuiltInStandardAttributes': {'members': [{'name': 'country-name',
                                                                          'optional': True,
                                                                          'type': 'CountryName'},
                                                                         {'name': 'administration-domain-name',
                                                                          'optional': True,
                                                                          'type': 'AdministrationDomainName'},
                                                                         {'name': 'network-address',
                                                                          'optional': True,
                                                                          'tag': {'kind': 'IMPLICIT',
                                                                                  'number': 0},
                                                                          'type': 'NetworkAddress'},
                                                                         {'name': 'terminal-identifier',
                                                                          'optional': True,
                                                                          'tag': {'kind': 'IMPLICIT',
                                                                                  'number': 1},
                                                                          'type': 'TerminalIdentifier'},
                                                                         {'name': 'private-domain-name',
                                                                          'optional': True,
                                                                          'tag': {'number': 2},
                                                                          'type': 'PrivateDomainName'},
                                                                         {'name': 'organization-name',
                                                                          'optional': True,
                                                                          'tag': {'kind': 'IMPLICIT',
                                                                                  'number': 3},
                                                                          'type': 'OrganizationName'},
                                                                         {'name': 'numeric-user-identifier',
                                                                          'optional': True,
                                                                          'tag': {'kind': 'IMPLICIT',
                                                                                  'number': 4},
                                                                          'type': 'NumericUserIdentifier'},
                                                                         {'name': 'personal-name',
                                                                          'optional': True,
                                                                          'tag': {'kind': 'IMPLICIT',
                                                                                  'number': 5},
                                                                          'type': 'PersonalName'},
                                                                         {'name': 'organizational-unit-names',
                                                                          'optional': True,
                                                                          'tag': {'kind': 'IMPLICIT',
                                                                                  'number': 6},
                                                                          'type': 'OrganizationalUnitNames'}],
                                                             'type': 'SEQUENCE'},
                               'Certificate': {'members': [{'name': 'tbsCertificate',
                                                            'type': 'TBSCertificate'},
                                                           {'name': 'signatureAlgorithm',
                                                            'type': 'AlgorithmIdentifier'},
                                                           {'name': 'signature',
                                                            'type': 'BIT '
                                                                    'STRING'}],
                                               'type': 'SEQUENCE'},
                               'CertificateList': {'members': [{'name': 'tbsCertList',
                                                                'type': 'TBSCertList'},
                                                               {'name': 'signatureAlgorithm',
                                                                'type': 'AlgorithmIdentifier'},
                                                               {'name': 'signature',
                                                                'type': 'BIT '
                                                                        'STRING'}],
                                                   'type': 'SEQUENCE'},
                               'CertificateSerialNumber': {'type': 'INTEGER'},
                               'CommonName': {'size': [(1,
                                                        'ub-common-name-length')],
                                              'type': 'PrintableString'},
                               'CountryName': {'members': [{'name': 'x121-dcc-code',
                                                            'size': ['ub-country-name-numeric-length'],
                                                            'type': 'NumericString'},
                                                           {'name': 'iso-3166-alpha2-code',
                                                            'size': ['ub-country-name-alpha-length'],
                                                            'type': 'PrintableString'}],
                                               'tag': {'class': 'APPLICATION',
                                                       'number': 1},
                                               'type': 'CHOICE'},
                               'DirectoryString': {'members': [{'name': 'teletexString',
                                                                'size': [(1,
                                                                          'MAX')],
                                                                'type': 'TeletexString'},
                                                               {'name': 'printableString',
                                                                'size': [(1,
                                                                          'MAX')],
                                                                'type': 'PrintableString'},
                                                               {'name': 'universalString',
                                                                'size': [(1,
                                                                          'MAX')],
                                                                'type': 'UniversalString'},
                                                               {'name': 'utf8String',
                                                                'size': [(1,
                                                                          'MAX')],
                                                                'type': 'UTF8String'},
                                                               {'name': 'bmpString',
                                                                'size': [(1,
                                                                          'MAX')],
                                                                'type': 'BMPString'}],
                                                   'type': 'CHOICE'},
                               'DistinguishedName': {'type': 'RDNSequence'},
                               'DomainComponent': {'type': 'IA5String'},
                               'EmailAddress': {'size': [(1,
                                                          'ub-emailaddress-length')],
                                                'type': 'IA5String'},
                               'ExtendedNetworkAddress': {'members': [{'members': [{'name': 'number',
                                                                                    'size': [(1,
                                                                                              'ub-e163-4-number-length')],
                                                                                    'tag': {'kind': 'IMPLICIT',
                                                                                            'number': 0},
                                                                                    'type': 'NumericString'},
                                                                                   {'name': 'sub-address',
                                                                                    'optional': True,
                                                                                    'size': [(1,
                                                                                              'ub-e163-4-sub-address-length')],
                                                                                    'tag': {'kind': 'IMPLICIT',
                                                                                            'number': 1},
                                                                                    'type': 'NumericString'}],
                                                                       'name': 'e163-4-address',
                                                                       'type': 'SEQUENCE'},
                                                                      {'name': 'psap-address',
                                                                       'tag': {'kind': 'IMPLICIT',
                                                                               'number': 0},
                                                                       'type': 'PresentationAddress'}],
                                                          'type': 'CHOICE'},
                               'Extension': {'members': [{'name': 'extnID',
                                                          'type': 'OBJECT '
                                                                  'IDENTIFIER'},
                                                         {'default': False,
                                                          'name': 'critical',
                                                          'type': 'BOOLEAN'},
                                                         {'name': 'extnValue',
                                                          'type': 'OCTET '
                                                                  'STRING'}],
                                             'type': 'SEQUENCE'},
                               'ExtensionAttribute': {'members': [{'name': 'extension-attribute-type',
                                                                   'restricted-to': [(0,
                                                                                      'ub-extension-attributes')],
                                                                   'tag': {'kind': 'IMPLICIT',
                                                                           'number': 0},
                                                                   'type': 'INTEGER'},
                                                                  {'choices': {},
                                                                   'name': 'extension-attribute-value',
                                                                   'tag': {'number': 1},
                                                                   'type': 'ANY '
                                                                           'DEFINED '
                                                                           'BY',
                                                                   'value': 'extension-attribute-type'}],
                                                      'type': 'SEQUENCE'},
                               'ExtensionAttributes': {'element': {'type': 'ExtensionAttribute'},
                                                       'size': [(1,
                                                                 'ub-extension-attributes')],
                                                       'type': 'SET OF'},
                               'ExtensionORAddressComponents': {'type': 'PDSParameter'},
                               'ExtensionPhysicalDeliveryAddressComponents': {'type': 'PDSParameter'},
                               'Extensions': {'element': {'type': 'Extension'},
                                              'size': [(1, 'MAX')],
                                              'type': 'SEQUENCE OF'},
                               'LocalPostalAttributes': {'type': 'PDSParameter'},
                               'Name': {'members': [{'name': 'rdnSequence',
                                                     'type': 'RDNSequence'}],
                                        'type': 'CHOICE'},
                               'NetworkAddress': {'type': 'X121Address'},
                               'NumericUserIdentifier': {'size': [(1,
                                                                   'ub-numeric-user-id-length')],
                                                         'type': 'NumericString'},
                               'ORAddress': {'members': [{'name': 'built-in-standard-attributes',
                                                          'type': 'BuiltInStandardAttributes'},
                                                         {'name': 'built-in-domain-defined-attributes',
                                                          'optional': True,
                                                          'type': 'BuiltInDomainDefinedAttributes'},
                                                         {'name': 'extension-attributes',
                                                          'optional': True,
                                                          'type': 'ExtensionAttributes'}],
                                             'type': 'SEQUENCE'},
                               'OrganizationName': {'size': [(1,
                                                              'ub-organization-name-length')],
                                                    'type': 'PrintableString'},
                               'OrganizationalUnitName': {'size': [(1,
                                                                    'ub-organizational-unit-name-length')],
                                                          'type': 'PrintableString'},
                               'OrganizationalUnitNames': {'element': {'type': 'OrganizationalUnitName'},
                                                           'size': [(1,
                                                                     'ub-organizational-units')],
                                                           'type': 'SEQUENCE '
                                                                   'OF'},
                               'PDSName': {'size': [(1, 'ub-pds-name-length')],
                                           'type': 'PrintableString'},
                               'PDSParameter': {'members': [{'name': 'printable-string',
                                                             'optional': True,
                                                             'size': [(1,
                                                                       'ub-pds-parameter-length')],
                                                             'type': 'PrintableString'},
                                                            {'name': 'teletex-string',
                                                             'optional': True,
                                                             'size': [(1,
                                                                       'ub-pds-parameter-length')],
                                                             'type': 'TeletexString'}],
                                                'type': 'SET'},
                               'PersonalName': {'members': [{'name': 'surname',
                                                             'size': [(1,
                                                                       'ub-surname-length')],
                                                             'tag': {'kind': 'IMPLICIT',
                                                                     'number': 0},
                                                             'type': 'PrintableString'},
                                                            {'name': 'given-name',
                                                             'optional': True,
                                                             'size': [(1,
                                                                       'ub-given-name-length')],
                                                             'tag': {'kind': 'IMPLICIT',
                                                                     'number': 1},
                                                             'type': 'PrintableString'},
                                                            {'name': 'initials',
                                                             'optional': True,
                                                             'size': [(1,
                                                                       'ub-initials-length')],
                                                             'tag': {'kind': 'IMPLICIT',
                                                                     'number': 2},
                                                             'type': 'PrintableString'},
                                                            {'name': 'generation-qualifier',
                                                             'optional': True,
                                                             'size': [(1,
                                                                       'ub-generation-qualifier-length')],
                                                             'tag': {'kind': 'IMPLICIT',
                                                                     'number': 3},
                                                             'type': 'PrintableString'}],
                                                'type': 'SET'},
                               'PhysicalDeliveryCountryName': {'members': [{'name': 'x121-dcc-code',
                                                                            'size': ['ub-country-name-numeric-length'],
                                                                            'type': 'NumericString'},
                                                                           {'name': 'iso-3166-alpha2-code',
                                                                            'size': ['ub-country-name-alpha-length'],
                                                                            'type': 'PrintableString'}],
                                                               'type': 'CHOICE'},
                               'PhysicalDeliveryOfficeName': {'type': 'PDSParameter'},
                               'PhysicalDeliveryOfficeNumber': {'type': 'PDSParameter'},
                               'PhysicalDeliveryOrganizationName': {'type': 'PDSParameter'},
                               'PhysicalDeliveryPersonalName': {'type': 'PDSParameter'},
                               'PostOfficeBoxAddress': {'type': 'PDSParameter'},
                               'PostalCode': {'members': [{'name': 'numeric-code',
                                                           'size': [(1,
                                                                     'ub-postal-code-length')],
                                                           'type': 'NumericString'},
                                                          {'name': 'printable-code',
                                                           'size': [(1,
                                                                     'ub-postal-code-length')],
                                                           'type': 'PrintableString'}],
                                              'type': 'CHOICE'},
                               'PosteRestanteAddress': {'type': 'PDSParameter'},
                               'PresentationAddress': {'members': [{'name': 'pSelector',
                                                                    'optional': True,
                                                                    'tag': {'kind': 'EXPLICIT',
                                                                            'number': 0},
                                                                    'type': 'OCTET '
                                                                            'STRING'},
                                                                   {'name': 'sSelector',
                                                                    'optional': True,
                                                                    'tag': {'kind': 'EXPLICIT',
                                                                            'number': 1},
                                                                    'type': 'OCTET '
                                                                            'STRING'},
                                                                   {'name': 'tSelector',
                                                                    'optional': True,
                                                                    'tag': {'kind': 'EXPLICIT',
                                                                            'number': 2},
                                                                    'type': 'OCTET '
                                                                            'STRING'},
                                                                   {'element': {'type': 'OCTET '
                                                                                        'STRING'},
                                                                    'name': 'nAddresses',
                                                                    'size': [(1,
                                                                              'MAX')],
                                                                    'tag': {'kind': 'EXPLICIT',
                                                                            'number': 3},
                                                                    'type': 'SET '
                                                                            'OF'}],
                                                       'type': 'SEQUENCE'},
                               'PrivateDomainName': {'members': [{'name': 'numeric',
                                                                  'size': [(1,
                                                                            'ub-domain-name-length')],
                                                                  'type': 'NumericString'},
                                                                 {'name': 'printable',
                                                                  'size': [(1,
                                                                            'ub-domain-name-length')],
                                                                  'type': 'PrintableString'}],
                                                     'type': 'CHOICE'},
                               'RDNSequence': {'element': {'type': 'RelativeDistinguishedName'},
                                               'type': 'SEQUENCE OF'},
                               'RelativeDistinguishedName': {'element': {'type': 'AttributeTypeAndValue'},
                                                             'size': [(1,
                                                                       'MAX')],
                                                             'type': 'SET OF'},
                               'StreetAddress': {'type': 'PDSParameter'},
                               'SubjectPublicKeyInfo': {'members': [{'name': 'algorithm',
                                                                     'type': 'AlgorithmIdentifier'},
                                                                    {'name': 'subjectPublicKey',
                                                                     'type': 'BIT '
                                                                             'STRING'}],
                                                        'type': 'SEQUENCE'},
                               'TBSCertList': {'members': [{'name': 'version',
                                                            'optional': True,
                                                            'type': 'Version'},
                                                           {'name': 'signature',
                                                            'type': 'AlgorithmIdentifier'},
                                                           {'name': 'issuer',
                                                            'type': 'Name'},
                                                           {'name': 'thisUpdate',
                                                            'type': 'Time'},
                                                           {'name': 'nextUpdate',
                                                            'optional': True,
                                                            'type': 'Time'},
                                                           {'element': {'members': [{'name': 'userCertificate',
                                                                                     'type': 'CertificateSerialNumber'},
                                                                                    {'name': 'revocationDate',
                                                                                     'type': 'Time'},
                                                                                    {'name': 'crlEntryExtensions',
                                                                                     'optional': True,
                                                                                     'type': 'Extensions'}],
                                                                        'type': 'SEQUENCE'},
                                                            'name': 'revokedCertificates',
                                                            'optional': True,
                                                            'type': 'SEQUENCE '
                                                                    'OF'},
                                                           {'name': 'crlExtensions',
                                                            'optional': True,
                                                            'tag': {'number': 0},
                                                            'type': 'Extensions'}],
                                               'type': 'SEQUENCE'},
                               'TBSCertificate': {'members': [{'default': 'v1',
                                                               'name': 'version',
                                                               'tag': {'number': 0},
                                                               'type': 'Version'},
                                                              {'name': 'serialNumber',
                                                               'type': 'CertificateSerialNumber'},
                                                              {'name': 'signature',
                                                               'type': 'AlgorithmIdentifier'},
                                                              {'name': 'issuer',
                                                               'type': 'Name'},
                                                              {'name': 'validity',
                                                               'type': 'Validity'},
                                                              {'name': 'subject',
                                                               'type': 'Name'},
                                                              {'name': 'subjectPublicKeyInfo',
                                                               'type': 'SubjectPublicKeyInfo'},
                                                              {'name': 'issuerUniqueID',
                                                               'optional': True,
                                                               'tag': {'kind': 'IMPLICIT',
                                                                       'number': 1},
                                                               'type': 'UniqueIdentifier'},
                                                              {'name': 'subjectUniqueID',
                                                               'optional': True,
                                                               'tag': {'kind': 'IMPLICIT',
                                                                       'number': 2},
                                                               'type': 'UniqueIdentifier'},
                                                              {'name': 'extensions',
                                                               'optional': True,
                                                               'tag': {'number': 3},
                                                               'type': 'Extensions'}],
                                                  'type': 'SEQUENCE'},
                               'TeletexCommonName': {'size': [(1,
                                                               'ub-common-name-length')],
                                                     'type': 'TeletexString'},
                               'TeletexDomainDefinedAttribute': {'members': [{'name': 'type',
                                                                              'size': [(1,
                                                                                        'ub-domain-defined-attribute-type-length')],
                                                                              'type': 'TeletexString'},
                                                                             {'name': 'value',
                                                                              'size': [(1,
                                                                                        'ub-domain-defined-attribute-value-length')],
                                                                              'type': 'TeletexString'}],
                                                                 'type': 'SEQUENCE'},
                               'TeletexDomainDefinedAttributes': {'element': {'type': 'TeletexDomainDefinedAttribute'},
                                                                  'size': [(1,
                                                                            'ub-domain-defined-attributes')],
                                                                  'type': 'SEQUENCE '
                                                                          'OF'},
                               'TeletexOrganizationName': {'size': [(1,
                                                                     'ub-organization-name-length')],
                                                           'type': 'TeletexString'},
                               'TeletexOrganizationalUnitName': {'size': [(1,
                                                                           'ub-organizational-unit-name-length')],
                                                                 'type': 'TeletexString'},
                               'TeletexOrganizationalUnitNames': {'element': {'type': 'TeletexOrganizationalUnitName'},
                                                                  'size': [(1,
                                                                            'ub-organizational-units')],
                                                                  'type': 'SEQUENCE '
                                                                          'OF'},
                               'TeletexPersonalName': {'members': [{'name': 'surname',
                                                                    'size': [(1,
                                                                              'ub-surname-length')],
                                                                    'tag': {'kind': 'IMPLICIT',
                                                                            'number': 0},
                                                                    'type': 'TeletexString'},
                                                                   {'name': 'given-name',
                                                                    'optional': True,
                                                                    'size': [(1,
                                                                              'ub-given-name-length')],
                                                                    'tag': {'kind': 'IMPLICIT',
                                                                            'number': 1},
                                                                    'type': 'TeletexString'},
                                                                   {'name': 'initials',
                                                                    'optional': True,
                                                                    'size': [(1,
                                                                              'ub-initials-length')],
                                                                    'tag': {'kind': 'IMPLICIT',
                                                                            'number': 2},
                                                                    'type': 'TeletexString'},
                                                                   {'name': 'generation-qualifier',
                                                                    'optional': True,
                                                                    'size': [(1,
                                                                              'ub-generation-qualifier-length')],
                                                                    'tag': {'kind': 'IMPLICIT',
                                                                            'number': 3},
                                                                    'type': 'TeletexString'}],
                                                       'type': 'SET'},
                               'TerminalIdentifier': {'size': [(1,
                                                                'ub-terminal-id-length')],
                                                      'type': 'PrintableString'},
                               'TerminalType': {'restricted-to': [(0,
                                                                   'ub-integer-options')],
                                                'type': 'INTEGER'},
                               'Time': {'members': [{'name': 'utcTime',
                                                     'type': 'UTCTime'},
                                                    {'name': 'generalTime',
                                                     'type': 'GeneralizedTime'}],
                                        'type': 'CHOICE'},
                               'UnformattedPostalAddress': {'members': [{'element': {'size': [(1,
                                                                                               'ub-pds-parameter-length')],
                                                                                     'type': 'PrintableString'},
                                                                         'name': 'printable-address',
                                                                         'optional': True,
                                                                         'size': [(1,
                                                                                   'ub-pds-physical-address-lines')],
                                                                         'type': 'SEQUENCE '
                                                                                 'OF'},
                                                                        {'name': 'teletex-string',
                                                                         'optional': True,
                                                                         'size': [(1,
                                                                                   'ub-unformatted-address-length')],
                                                                         'type': 'TeletexString'}],
                                                            'type': 'SET'},
                               'UniqueIdentifier': {'type': 'BIT STRING'},
                               'UniquePostalName': {'type': 'PDSParameter'},
                               'Validity': {'members': [{'name': 'notBefore',
                                                         'type': 'Time'},
                                                        {'name': 'notAfter',
                                                         'type': 'Time'}],
                                            'type': 'SEQUENCE'},
                               'Version': {'type': 'INTEGER'},
                               'X121Address': {'size': [(1,
                                                         'ub-x121-address-length')],
                                               'type': 'NumericString'},
                               'X520CommonName': {'members': [{'name': 'teletexString',
                                                               'size': [(1,
                                                                         'ub-common-name')],
                                                               'type': 'TeletexString'},
                                                              {'name': 'printableString',
                                                               'size': [(1,
                                                                         'ub-common-name')],
                                                               'type': 'PrintableString'},
                                                              {'name': 'universalString',
                                                               'size': [(1,
                                                                         'ub-common-name')],
                                                               'type': 'UniversalString'},
                                                              {'name': 'utf8String',
                                                               'size': [(1,
                                                                         'ub-common-name')],
                                                               'type': 'UTF8String'},
                                                              {'name': 'bmpString',
                                                               'size': [(1,
                                                                         'ub-common-name')],
                                                               'type': 'BMPString'}],
                                                  'type': 'CHOICE'},
                               'X520LocalityName': {'members': [{'name': 'teletexString',
                                                                 'size': [(1,
                                                                           'ub-locality-name')],
                                                                 'type': 'TeletexString'},
                                                                {'name': 'printableString',
                                                                 'size': [(1,
                                                                           'ub-locality-name')],
                                                                 'type': 'PrintableString'},
                                                                {'name': 'universalString',
                                                                 'size': [(1,
                                                                           'ub-locality-name')],
                                                                 'type': 'UniversalString'},
                                                                {'name': 'utf8String',
                                                                 'size': [(1,
                                                                           'ub-locality-name')],
                                                                 'type': 'UTF8String'},
                                                                {'name': 'bmpString',
                                                                 'size': [(1,
                                                                           'ub-locality-name')],
                                                                 'type': 'BMPString'}],
                                                    'type': 'CHOICE'},
                               'X520OrganizationName': {'members': [{'name': 'teletexString',
                                                                     'size': [(1,
                                                                               'ub-organization-name')],
                                                                     'type': 'TeletexString'},
                                                                    {'name': 'printableString',
                                                                     'size': [(1,
                                                                               'ub-organization-name')],
                                                                     'type': 'PrintableString'},
                                                                    {'name': 'universalString',
                                                                     'size': [(1,
                                                                               'ub-organization-name')],
                                                                     'type': 'UniversalString'},
                                                                    {'name': 'utf8String',
                                                                     'size': [(1,
                                                                               'ub-organization-name')],
                                                                     'type': 'UTF8String'},
                                                                    {'name': 'bmpString',
                                                                     'size': [(1,
                                                                               'ub-organization-name')],
                                                                     'type': 'BMPString'}],
                                                        'type': 'CHOICE'},
                               'X520OrganizationalUnitName': {'members': [{'name': 'teletexString',
                                                                           'size': [(1,
                                                                                     'ub-organizational-unit-name')],
                                                                           'type': 'TeletexString'},
                                                                          {'name': 'printableString',
                                                                           'size': [(1,
                                                                                     'ub-organizational-unit-name')],
                                                                           'type': 'PrintableString'},
                                                                          {'name': 'universalString',
                                                                           'size': [(1,
                                                                                     'ub-organizational-unit-name')],
                                                                           'type': 'UniversalString'},
                                                                          {'name': 'utf8String',
                                                                           'size': [(1,
                                                                                     'ub-organizational-unit-name')],
                                                                           'type': 'UTF8String'},
                                                                          {'name': 'bmpString',
                                                                           'size': [(1,
                                                                                     'ub-organizational-unit-name')],
                                                                           'type': 'BMPString'}],
                                                              'type': 'CHOICE'},
                               'X520Pseudonym': {'members': [{'name': 'teletexString',
                                                              'size': [(1,
                                                                        'ub-pseudonym')],
                                                              'type': 'TeletexString'},
                                                             {'name': 'printableString',
                                                              'size': [(1,
                                                                        'ub-pseudonym')],
                                                              'type': 'PrintableString'},
                                                             {'name': 'universalString',
                                                              'size': [(1,
                                                                        'ub-pseudonym')],
                                                              'type': 'UniversalString'},
                                                             {'name': 'utf8String',
                                                              'size': [(1,
                                                                        'ub-pseudonym')],
                                                              'type': 'UTF8String'},
                                                             {'name': 'bmpString',
                                                              'size': [(1,
                                                                        'ub-pseudonym')],
                                                              'type': 'BMPString'}],
                                                 'type': 'CHOICE'},
                               'X520SerialNumber': {'size': [(1,
                                                              'ub-serial-number')],
                                                    'type': 'PrintableString'},
                               'X520StateOrProvinceName': {'members': [{'name': 'teletexString',
                                                                        'size': [(1,
                                                                                  'ub-state-name')],
                                                                        'type': 'TeletexString'},
                                                                       {'name': 'printableString',
                                                                        'size': [(1,
                                                                                  'ub-state-name')],
                                                                        'type': 'PrintableString'},
                                                                       {'name': 'universalString',
                                                                        'size': [(1,
                                                                                  'ub-state-name')],
                                                                        'type': 'UniversalString'},
                                                                       {'name': 'utf8String',
                                                                        'size': [(1,
                                                                                  'ub-state-name')],
                                                                        'type': 'UTF8String'},
                                                                       {'name': 'bmpString',
                                                                        'size': [(1,
                                                                                  'ub-state-name')],
                                                                        'type': 'BMPString'}],
                                                           'type': 'CHOICE'},
                               'X520Title': {'members': [{'name': 'teletexString',
                                                          'size': [(1,
                                                                    'ub-title')],
                                                          'type': 'TeletexString'},
                                                         {'name': 'printableString',
                                                          'size': [(1,
                                                                    'ub-title')],
                                                          'type': 'PrintableString'},
                                                         {'name': 'universalString',
                                                          'size': [(1,
                                                                    'ub-title')],
                                                          'type': 'UniversalString'},
                                                         {'name': 'utf8String',
                                                          'size': [(1,
                                                                    'ub-title')],
                                                          'type': 'UTF8String'},
                                                         {'name': 'bmpString',
                                                          'size': [(1,
                                                                    'ub-title')],
                                                          'type': 'BMPString'}],
                                             'type': 'CHOICE'},
                               'X520countryName': {'size': [2],
                                                   'type': 'PrintableString'},
                               'X520dnQualifier': {'type': 'PrintableString'},
                               'X520name': {'members': [{'name': 'teletexString',
                                                         'size': [(1,
                                                                   'ub-name')],
                                                         'type': 'TeletexString'},
                                                        {'name': 'printableString',
                                                         'size': [(1,
                                                                   'ub-name')],
                                                         'type': 'PrintableString'},
                                                        {'name': 'universalString',
                                                         'size': [(1,
                                                                   'ub-name')],
                                                         'type': 'UniversalString'},
                                                        {'name': 'utf8String',
                                                         'size': [(1,
                                                                   'ub-name')],
                                                         'type': 'UTF8String'},
                                                        {'name': 'bmpString',
                                                         'size': [(1,
                                                                   'ub-name')],
                                                         'type': 'BMPString'}],
                                            'type': 'CHOICE'}},
                     'values': {'common-name': {'type': 'INTEGER', 'value': 1},
                                'extended-network-address': {'type': 'INTEGER',
                                                             'value': 22},
                                'extension-OR-address-components': {'type': 'INTEGER',
                                                                    'value': 12},
                                'extension-physical-delivery-address-components': {'type': 'INTEGER',
                                                                                   'value': 15},
                                'id-ad': {'type': 'OBJECT IDENTIFIER',
                                          'value': ['id-pkix', 48]},
                                'id-ad-caIssuers': {'type': 'OBJECT IDENTIFIER',
                                                    'value': ['id-ad', 2]},
                                'id-ad-caRepository': {'type': 'OBJECT '
                                                               'IDENTIFIER',
                                                       'value': ['id-ad', 5]},
                                'id-ad-ocsp': {'type': 'OBJECT IDENTIFIER',
                                               'value': ['id-ad', 1]},
                                'id-ad-timeStamping': {'type': 'OBJECT '
                                                               'IDENTIFIER',
                                                       'value': ['id-ad', 3]},
                                'id-at': {'type': 'OBJECT IDENTIFIER',
                                          'value': [('joint-iso-ccitt', 2),
                                                    ('ds', 5),
                                                    4]},
                                'id-at-commonName': {'type': 'AttributeType',
                                                     'value': None},
                                'id-at-countryName': {'type': 'AttributeType',
                                                      'value': None},
                                'id-at-dnQualifier': {'type': 'AttributeType',
                                                      'value': None},
                                'id-at-generationQualifier': {'type': 'AttributeType',
                                                              'value': None},
                                'id-at-givenName': {'type': 'AttributeType',
                                                    'value': None},
                                'id-at-initials': {'type': 'AttributeType',
                                                   'value': None},
                                'id-at-localityName': {'type': 'AttributeType',
                                                       'value': None},
                                'id-at-name': {'type': 'AttributeType',
                                               'value': None},
                                'id-at-organizationName': {'type': 'AttributeType',
                                                           'value': None},
                                'id-at-organizationalUnitName': {'type': 'AttributeType',
                                                                 'value': None},
                                'id-at-pseudonym': {'type': 'AttributeType',
                                                    'value': None},
                                'id-at-serialNumber': {'type': 'AttributeType',
                                                       'value': None},
                                'id-at-stateOrProvinceName': {'type': 'AttributeType',
                                                              'value': None},
                                'id-at-surname': {'type': 'AttributeType',
                                                  'value': None},
                                'id-at-title': {'type': 'AttributeType',
                                                'value': None},
                                'id-domainComponent': {'type': 'AttributeType',
                                                       'value': None},
                                'id-emailAddress': {'type': 'AttributeType',
                                                    'value': None},
                                'id-kp': {'type': 'OBJECT IDENTIFIER',
                                          'value': ['id-pkix', 3]},
                                'id-pe': {'type': 'OBJECT IDENTIFIER',
                                          'value': ['id-pkix', 1]},
                                'id-pkix': {'type': 'OBJECT IDENTIFIER',
                                            'value': [('iso', 1),
                                                      ('identified-organization',
                                                       3),
                                                      ('dod', 6),
                                                      ('internet', 1),
                                                      ('security', 5),
                                                      ('mechanisms', 5),
                                                      ('pkix', 7)]},
                                'id-qt': {'type': 'OBJECT IDENTIFIER',
                                          'value': ['id-pkix', 2]},
                                'id-qt-cps': {'type': 'OBJECT IDENTIFIER',
                                              'value': ['id-qt', 1]},
                                'id-qt-unotice': {'type': 'OBJECT IDENTIFIER',
                                                  'value': ['id-qt', 2]},
                                'local-postal-attributes': {'type': 'INTEGER',
                                                            'value': 21},
                                'pds-name': {'type': 'INTEGER', 'value': 7},
                                'physical-delivery-country-name': {'type': 'INTEGER',
                                                                   'value': 8},
                                'physical-delivery-office-name': {'type': 'INTEGER',
                                                                  'value': 10},
                                'physical-delivery-office-number': {'type': 'INTEGER',
                                                                    'value': 11},
                                'physical-delivery-organization-name': {'type': 'INTEGER',
                                                                        'value': 14},
                                'physical-delivery-personal-name': {'type': 'INTEGER',
                                                                    'value': 13},
                                'pkcs-9': {'type': 'OBJECT IDENTIFIER',
                                           'value': [('iso', 1),
                                                     ('member-body', 2),
                                                     ('us', 840),
                                                     ('rsadsi', 113549),
                                                     ('pkcs', 1),
                                                     9]},
                                'post-office-box-address': {'type': 'INTEGER',
                                                            'value': 18},
                                'postal-code': {'type': 'INTEGER', 'value': 9},
                                'poste-restante-address': {'type': 'INTEGER',
                                                           'value': 19},
                                'street-address': {'type': 'INTEGER',
                                                   'value': 17},
                                'teletex-common-name': {'type': 'INTEGER',
                                                        'value': 2},
                                'teletex-domain-defined-attributes': {'type': 'INTEGER',
                                                                      'value': 6},
                                'teletex-organization-name': {'type': 'INTEGER',
                                                              'value': 3},
                                'teletex-organizational-unit-names': {'type': 'INTEGER',
                                                                      'value': 5},
                                'teletex-personal-name': {'type': 'INTEGER',
                                                          'value': 4},
                                'terminal-type': {'type': 'INTEGER',
                                                  'value': 23},
                                'ub-common-name': {'type': 'INTEGER',
                                                   'value': 64},
                                'ub-common-name-length': {'type': 'INTEGER',
                                                          'value': 64},
                                'ub-country-name-alpha-length': {'type': 'INTEGER',
                                                                 'value': 2},
                                'ub-country-name-numeric-length': {'type': 'INTEGER',
                                                                   'value': 3},
                                'ub-domain-defined-attribute-type-length': {'type': 'INTEGER',
                                                                            'value': 8},
                                'ub-domain-defined-attribute-value-length': {'type': 'INTEGER',
                                                                             'value': 128},
                                'ub-domain-defined-attributes': {'type': 'INTEGER',
                                                                 'value': 4},
                                'ub-domain-name-length': {'type': 'INTEGER',
                                                          'value': 16},
                                'ub-e163-4-number-length': {'type': 'INTEGER',
                                                            'value': 15},
                                'ub-e163-4-sub-address-length': {'type': 'INTEGER',
                                                                 'value': 40},
                                'ub-emailaddress-length': {'type': 'INTEGER',
                                                           'value': 255},
                                'ub-extension-attributes': {'type': 'INTEGER',
                                                            'value': 256},
                                'ub-generation-qualifier-length': {'type': 'INTEGER',
                                                                   'value': 3},
                                'ub-given-name-length': {'type': 'INTEGER',
                                                         'value': 16},
                                'ub-initials-length': {'type': 'INTEGER',
                                                       'value': 5},
                                'ub-integer-options': {'type': 'INTEGER',
                                                       'value': 256},
                                'ub-locality-name': {'type': 'INTEGER',
                                                     'value': 128},
                                'ub-match': {'type': 'INTEGER', 'value': 128},
                                'ub-name': {'type': 'INTEGER', 'value': 32768},
                                'ub-numeric-user-id-length': {'type': 'INTEGER',
                                                              'value': 32},
                                'ub-organization-name': {'type': 'INTEGER',
                                                         'value': 64},
                                'ub-organization-name-length': {'type': 'INTEGER',
                                                                'value': 64},
                                'ub-organizational-unit-name': {'type': 'INTEGER',
                                                                'value': 64},
                                'ub-organizational-unit-name-length': {'type': 'INTEGER',
                                                                       'value': 32},
                                'ub-organizational-units': {'type': 'INTEGER',
                                                            'value': 4},
                                'ub-pds-name-length': {'type': 'INTEGER',
                                                       'value': 16},
                                'ub-pds-parameter-length': {'type': 'INTEGER',
                                                            'value': 30},
                                'ub-pds-physical-address-lines': {'type': 'INTEGER',
                                                                  'value': 6},
                                'ub-postal-code-length': {'type': 'INTEGER',
                                                          'value': 16},
                                'ub-pseudonym': {'type': 'INTEGER',
                                                 'value': 128},
                                'ub-serial-number': {'type': 'INTEGER',
                                                     'value': 64},
                                'ub-state-name': {'type': 'INTEGER',
                                                  'value': 128},
                                'ub-surname-length': {'type': 'INTEGER',
                                                      'value': 40},
                                'ub-terminal-id-length': {'type': 'INTEGER',
                                                          'value': 24},
                                'ub-title': {'type': 'INTEGER', 'value': 64},
                                'ub-unformatted-address-length': {'type': 'INTEGER',
                                                                  'value': 180},
                                'ub-x121-address-length': {'type': 'INTEGER',
                                                           'value': 16},
                                'unformatted-postal-address': {'type': 'INTEGER',
                                                               'value': 16},
                                'unique-postal-name': {'type': 'INTEGER',
                                                       'value': 20}}},
 'PKIX1Implicit88': {'extensibility-implied': False,
                     'imports': {'PKIX1Explicit88': ['id-pe',
                                                     'id-kp',
                                                     'id-qt-unotice',
                                                     'id-qt-cps',
                                                     'BMPString',
                                                     'UTF8String',
                                                     'ORAddress',
                                                     'Name',
                                                     'RelativeDistinguishedName',
                                                     'CertificateSerialNumber',
                                                     'Attribute',
                                                     'DirectoryString']},
                     'object-classes': {},
                     'object-sets': {},
                     'tags': 'IMPLICIT',
                     'types': {'AccessDescription': {'members': [{'name': 'accessMethod',
                                                                  'type': 'OBJECT '
                                                                          'IDENTIFIER'},
                                                                 {'name': 'accessLocation',
                                                                  'type': 'GeneralName'}],
                                                     'type': 'SEQUENCE'},
                               'AnotherName': {'members': [{'name': 'type-id',
                                                            'type': 'OBJECT '
                                                                    'IDENTIFIER'},
                                                           {'choices': {},
                                                            'name': 'value',
                                                            'tag': {'kind': 'EXPLICIT',
                                                                    'number': 0},
                                                            'type': 'ANY '
                                                                    'DEFINED '
                                                                    'BY',
                                                            'value': 'type-id'}],
                                               'type': 'SEQUENCE'},
                               'AuthorityInfoAccessSyntax': {'element': {'type': 'AccessDescription'},
                                                             'size': [(1,
                                                                       'MAX')],
                                                             'type': 'SEQUENCE '
                                                                     'OF'},
                               'AuthorityKeyIdentifier': {'members': [{'name': 'keyIdentifier',
                                                                       'optional': True,
                                                                       'tag': {'number': 0},
                                                                       'type': 'KeyIdentifier'},
                                                                      {'name': 'authorityCertIssuer',
                                                                       'optional': True,
                                                                       'tag': {'number': 1},
                                                                       'type': 'GeneralNames'},
                                                                      {'name': 'authorityCertSerialNumber',
                                                                       'optional': True,
                                                                       'tag': {'number': 2},
                                                                       'type': 'CertificateSerialNumber'}],
                                                          'type': 'SEQUENCE'},
                               'BaseCRLNumber': {'type': 'CRLNumber'},
                               'BaseDistance': {'restricted-to': [(0, 'MAX')],
                                                'type': 'INTEGER'},
                               'BasicConstraints': {'members': [{'default': False,
                                                                 'name': 'cA',
                                                                 'type': 'BOOLEAN'},
                                                                {'name': 'pathLenConstraint',
                                                                 'optional': True,
                                                                 'restricted-to': [(0,
                                                                                    'MAX')],
                                                                 'type': 'INTEGER'}],
                                                    'type': 'SEQUENCE'},
                               'CPSuri': {'type': 'IA5String'},
                               'CRLDistributionPoints': {'element': {'type': 'DistributionPoint'},
                                                         'size': [(1, 'MAX')],
                                                         'type': 'SEQUENCE OF'},
                               'CRLNumber': {'restricted-to': [(0, 'MAX')],
                                             'type': 'INTEGER'},
                               'CRLReason': {'type': 'ENUMERATED',
                                             'values': [('unspecified', 0),
                                                        ('keyCompromise', 1),
                                                        ('cACompromise', 2),
                                                        ('affiliationChanged',
                                                         3),
                                                        ('superseded', 4),
                                                        ('cessationOfOperation',
                                                         5),
                                                        ('certificateHold', 6),
                                                        ('removeFromCRL', 8),
                                                        ('privilegeWithdrawn',
                                                         9),
                                                        ('aACompromise', 10)]},
                               'CertPolicyId': {'type': 'OBJECT IDENTIFIER'},
                               'CertificateIssuer': {'type': 'GeneralNames'},
                               'CertificatePolicies': {'element': {'type': 'PolicyInformation'},
                                                       'size': [(1, 'MAX')],
                                                       'type': 'SEQUENCE OF'},
                               'DisplayText': {'members': [{'name': 'ia5String',
                                                            'size': [(1, 200)],
                                                            'type': 'IA5String'},
                                                           {'name': 'visibleString',
                                                            'size': [(1, 200)],
                                                            'type': 'VisibleString'},
                                                           {'name': 'bmpString',
                                                            'size': [(1, 200)],
                                                            'type': 'BMPString'},
                                                           {'name': 'utf8String',
                                                            'size': [(1, 200)],
                                                            'type': 'UTF8String'}],
                                               'type': 'CHOICE'},
                               'DistributionPoint': {'members': [{'name': 'distributionPoint',
                                                                  'optional': True,
                                                                  'tag': {'number': 0},
                                                                  'type': 'DistributionPointName'},
                                                                 {'name': 'reasons',
                                                                  'optional': True,
                                                                  'tag': {'number': 1},
                                                                  'type': 'ReasonFlags'},
                                                                 {'name': 'cRLIssuer',
                                                                  'optional': True,
                                                                  'tag': {'number': 2},
                                                                  'type': 'GeneralNames'}],
                                                     'type': 'SEQUENCE'},
                               'DistributionPointName': {'members': [{'name': 'fullName',
                                                                      'tag': {'number': 0},
                                                                      'type': 'GeneralNames'},
                                                                     {'name': 'nameRelativeToCRLIssuer',
                                                                      'tag': {'number': 1},
                                                                      'type': 'RelativeDistinguishedName'}],
                                                         'type': 'CHOICE'},
                               'EDIPartyName': {'members': [{'name': 'nameAssigner',
                                                             'optional': True,
                                                             'tag': {'number': 0},
                                                             'type': 'DirectoryString'},
                                                            {'name': 'partyName',
                                                             'tag': {'number': 1},
                                                             'type': 'DirectoryString'}],
                                                'type': 'SEQUENCE'},
                               'ExtKeyUsageSyntax': {'element': {'type': 'KeyPurposeId'},
                                                     'size': [(1, 'MAX')],
                                                     'type': 'SEQUENCE OF'},
                               'FreshestCRL': {'type': 'CRLDistributionPoints'},
                               'GeneralName': {'members': [{'name': 'otherName',
                                                            'tag': {'number': 0},
                                                            'type': 'AnotherName'},
                                                           {'name': 'rfc822Name',
                                                            'tag': {'number': 1},
                                                            'type': 'IA5String'},
                                                           {'name': 'dNSName',
                                                            'tag': {'number': 2},
                                                            'type': 'IA5String'},
                                                           {'name': 'x400Address',
                                                            'tag': {'number': 3},
                                                            'type': 'ORAddress'},
                                                           {'name': 'directoryName',
                                                            'tag': {'number': 4},
                                                            'type': 'Name'},
                                                           {'name': 'ediPartyName',
                                                            'tag': {'number': 5},
                                                            'type': 'EDIPartyName'},
                                                           {'name': 'uniformResourceIdentifier',
                                                            'tag': {'number': 6},
                                                            'type': 'IA5String'},
                                                           {'name': 'iPAddress',
                                                            'tag': {'number': 7},
                                                            'type': 'OCTET '
                                                                    'STRING'},
                                                           {'name': 'registeredID',
                                                            'tag': {'number': 8},
                                                            'type': 'OBJECT '
                                                                    'IDENTIFIER'}],
                                               'type': 'CHOICE'},
                               'GeneralNames': {'element': {'type': 'GeneralName'},
                                                'size': [(1, 'MAX')],
                                                'type': 'SEQUENCE OF'},
                               'GeneralSubtree': {'members': [{'name': 'base',
                                                               'type': 'GeneralName'},
                                                              {'default': 0,
                                                               'name': 'minimum',
                                                               'tag': {'number': 0},
                                                               'type': 'BaseDistance'},
                                                              {'name': 'maximum',
                                                               'optional': True,
                                                               'tag': {'number': 1},
                                                               'type': 'BaseDistance'}],
                                                  'type': 'SEQUENCE'},
                               'GeneralSubtrees': {'element': {'type': 'GeneralSubtree'},
                                                   'size': [(1, 'MAX')],
                                                   'type': 'SEQUENCE OF'},
                               'HoldInstructionCode': {'type': 'OBJECT '
                                                               'IDENTIFIER'},
                               'InhibitAnyPolicy': {'type': 'SkipCerts'},
                               'InvalidityDate': {'type': 'GeneralizedTime'},
                               'IssuerAltName': {'type': 'GeneralNames'},
                               'IssuingDistributionPoint': {'members': [{'name': 'distributionPoint',
                                                                         'optional': True,
                                                                         'tag': {'number': 0},
                                                                         'type': 'DistributionPointName'},
                                                                        {'default': False,
                                                                         'name': 'onlyContainsUserCerts',
                                                                         'tag': {'number': 1},
                                                                         'type': 'BOOLEAN'},
                                                                        {'default': False,
                                                                         'name': 'onlyContainsCACerts',
                                                                         'tag': {'number': 2},
                                                                         'type': 'BOOLEAN'},
                                                                        {'name': 'onlySomeReasons',
                                                                         'optional': True,
                                                                         'tag': {'number': 3},
                                                                         'type': 'ReasonFlags'},
                                                                        {'default': False,
                                                                         'name': 'indirectCRL',
                                                                         'tag': {'number': 4},
                                                                         'type': 'BOOLEAN'},
                                                                        {'default': False,
                                                                         'name': 'onlyContainsAttributeCerts',
                                                                         'tag': {'number': 5},
                                                                         'type': 'BOOLEAN'}],
                                                            'type': 'SEQUENCE'},
                               'KeyIdentifier': {'type': 'OCTET STRING'},
                               'KeyPurposeId': {'type': 'OBJECT IDENTIFIER'},
                               'KeyUsage': {'named-bits': [('digitalSignature',
                                                            '0'),
                                                           ('nonRepudiation',
                                                            '1'),
                                                           ('keyEncipherment',
                                                            '2'),
                                                           ('dataEncipherment',
                                                            '3'),
                                                           ('keyAgreement',
                                                            '4'),
                                                           ('keyCertSign', '5'),
                                                           ('cRLSign', '6'),
                                                           ('encipherOnly',
                                                            '7'),
                                                           ('decipherOnly',
                                                            '8')],
                                            'type': 'BIT STRING'},
                               'NameConstraints': {'members': [{'name': 'permittedSubtrees',
                                                                'optional': True,
                                                                'tag': {'number': 0},
                                                                'type': 'GeneralSubtrees'},
                                                               {'name': 'excludedSubtrees',
                                                                'optional': True,
                                                                'tag': {'number': 1},
                                                                'type': 'GeneralSubtrees'}],
                                                   'type': 'SEQUENCE'},
                               'NoticeReference': {'members': [{'name': 'organization',
                                                                'type': 'DisplayText'},
                                                               {'element': {'type': 'INTEGER'},
                                                                'name': 'noticeNumbers',
                                                                'type': 'SEQUENCE '
                                                                        'OF'}],
                                                   'type': 'SEQUENCE'},
                               'PolicyConstraints': {'members': [{'name': 'requireExplicitPolicy',
                                                                  'optional': True,
                                                                  'tag': {'number': 0},
                                                                  'type': 'SkipCerts'},
                                                                 {'name': 'inhibitPolicyMapping',
                                                                  'optional': True,
                                                                  'tag': {'number': 1},
                                                                  'type': 'SkipCerts'}],
                                                     'type': 'SEQUENCE'},
                               'PolicyInformation': {'members': [{'name': 'policyIdentifier',
                                                                  'type': 'CertPolicyId'},
                                                                 {'element': {'type': 'PolicyQualifierInfo'},
                                                                  'name': 'policyQualifiers',
                                                                  'optional': True,
                                                                  'size': [(1,
                                                                            'MAX')],
                                                                  'type': 'SEQUENCE '
                                                                          'OF'}],
                                                     'type': 'SEQUENCE'},
                               'PolicyMappings': {'element': {'members': [{'name': 'issuerDomainPolicy',
                                                                           'type': 'CertPolicyId'},
                                                                          {'name': 'subjectDomainPolicy',
                                                                           'type': 'CertPolicyId'}],
                                                              'type': 'SEQUENCE'},
                                                  'size': [(1, 'MAX')],
                                                  'type': 'SEQUENCE OF'},
                               'PolicyQualifierId': {'type': 'OBJECT '
                                                             'IDENTIFIER'},
                               'PolicyQualifierInfo': {'members': [{'name': 'policyQualifierId',
                                                                    'type': 'PolicyQualifierId'},
                                                                   {'choices': {},
                                                                    'name': 'qualifier',
                                                                    'type': 'ANY '
                                                                            'DEFINED '
                                                                            'BY',
                                                                    'value': 'policyQualifierId'}],
                                                       'type': 'SEQUENCE'},
                               'PrivateKeyUsagePeriod': {'members': [{'name': 'notBefore',
                                                                      'optional': True,
                                                                      'tag': {'number': 0},
                                                                      'type': 'GeneralizedTime'},
                                                                     {'name': 'notAfter',
                                                                      'optional': True,
                                                                      'tag': {'number': 1},
                                                                      'type': 'GeneralizedTime'}],
                                                         'type': 'SEQUENCE'},
                               'ReasonFlags': {'named-bits': [('unused', '0'),
                                                              ('keyCompromise',
                                                               '1'),
                                                              ('cACompromise',
                                                               '2'),
                                                              ('affiliationChanged',
                                                               '3'),
                                                              ('superseded',
                                                               '4'),
                                                              ('cessationOfOperation',
                                                               '5'),
                                                              ('certificateHold',
                                                               '6'),
                                                              ('privilegeWithdrawn',
                                                               '7'),
                                                              ('aACompromise',
                                                               '8')],
                                               'type': 'BIT STRING'},
                               'SkipCerts': {'restricted-to': [(0, 'MAX')],
                                             'type': 'INTEGER'},
                               'SubjectAltName': {'type': 'GeneralNames'},
                               'SubjectDirectoryAttributes': {'element': {'type': 'Attribute'},
                                                              'size': [(1,
                                                                        'MAX')],
                                                              'type': 'SEQUENCE '
                                                                      'OF'},
                               'SubjectInfoAccessSyntax': {'element': {'type': 'AccessDescription'},
                                                           'size': [(1, 'MAX')],
                                                           'type': 'SEQUENCE '
                                                                   'OF'},
                               'SubjectKeyIdentifier': {'type': 'KeyIdentifier'},
                               'UserNotice': {'members': [{'name': 'noticeRef',
                                                           'optional': True,
                                                           'type': 'NoticeReference'},
                                                          {'name': 'explicitText',
                                                           'optional': True,
                                                           'type': 'DisplayText'}],
                                              'type': 'SEQUENCE'}},
                     'values': {'anyExtendedKeyUsage': {'type': 'OBJECT '
                                                                'IDENTIFIER',
                                                        'value': ['id-ce-extKeyUsage',
                                                                  0]},
                                'anyPolicy': {'type': 'OBJECT IDENTIFIER',
                                              'value': ['id-ce-certificatePolicies',
                                                        0]},
                                'holdInstruction': {'type': 'OBJECT IDENTIFIER',
                                                    'value': [('joint-iso-itu-t',
                                                               2),
                                                              ('member-body',
                                                               2),
                                                              ('us', 840),
                                                              ('x9cm', 10040),
                                                              2]},
                                'id-ce': {'type': 'OBJECT IDENTIFIER',
                                          'value': [('joint-iso-ccitt', 2),
                                                    ('ds', 5),
                                                    29]},
                                'id-ce-authorityKeyIdentifier': {'type': 'OBJECT '
                                                                         'IDENTIFIER',
                                                                 'value': ['id-ce',
                                                                           35]},
                                'id-ce-basicConstraints': {'type': 'OBJECT '
                                                                   'IDENTIFIER',
                                                           'value': ['id-ce',
                                                                     19]},
                                'id-ce-cRLDistributionPoints': {'type': 'OBJECT '
                                                                        'IDENTIFIER',
                                                                'value': ['id-ce',
                                                                          31]},
                                'id-ce-cRLNumber': {'type': 'OBJECT IDENTIFIER',
                                                    'value': ['id-ce', 20]},
                                'id-ce-cRLReasons': {'type': 'OBJECT '
                                                             'IDENTIFIER',
                                                     'value': ['id-ce', 21]},
                                'id-ce-certificateIssuer': {'type': 'OBJECT '
                                                                    'IDENTIFIER',
                                                            'value': ['id-ce',
                                                                      29]},
                                'id-ce-certificatePolicies': {'type': 'OBJECT '
                                                                      'IDENTIFIER',
                                                              'value': ['id-ce',
                                                                        32]},
                                'id-ce-deltaCRLIndicator': {'type': 'OBJECT '
                                                                    'IDENTIFIER',
                                                            'value': ['id-ce',
                                                                      27]},
                                'id-ce-extKeyUsage': {'type': 'OBJECT '
                                                              'IDENTIFIER',
                                                      'value': ['id-ce', 37]},
                                'id-ce-freshestCRL': {'type': 'OBJECT '
                                                              'IDENTIFIER',
                                                      'value': ['id-ce', 46]},
                                'id-ce-holdInstructionCode': {'type': 'OBJECT '
                                                                      'IDENTIFIER',
                                                              'value': ['id-ce',
                                                                        23]},
                                'id-ce-inhibitAnyPolicy': {'type': 'OBJECT '
                                                                   'IDENTIFIER',
                                                           'value': ['id-ce',
                                                                     54]},
                                'id-ce-invalidityDate': {'type': 'OBJECT '
                                                                 'IDENTIFIER',
                                                         'value': ['id-ce',
                                                                   24]},
                                'id-ce-issuerAltName': {'type': 'OBJECT '
                                                                'IDENTIFIER',
                                                        'value': ['id-ce', 18]},
                                'id-ce-issuingDistributionPoint': {'type': 'OBJECT '
                                                                           'IDENTIFIER',
                                                                   'value': ['id-ce',
                                                                             28]},
                                'id-ce-keyUsage': {'type': 'OBJECT IDENTIFIER',
                                                   'value': ['id-ce', 15]},
                                'id-ce-nameConstraints': {'type': 'OBJECT '
                                                                  'IDENTIFIER',
                                                          'value': ['id-ce',
                                                                    30]},
                                'id-ce-policyConstraints': {'type': 'OBJECT '
                                                                    'IDENTIFIER',
                                                            'value': ['id-ce',
                                                                      36]},
                                'id-ce-policyMappings': {'type': 'OBJECT '
                                                                 'IDENTIFIER',
                                                         'value': ['id-ce',
                                                                   33]},
                                'id-ce-privateKeyUsagePeriod': {'type': 'OBJECT '
                                                                        'IDENTIFIER',
                                                                'value': ['id-ce',
                                                                          16]},
                                'id-ce-subjectAltName': {'type': 'OBJECT '
                                                                 'IDENTIFIER',
                                                         'value': ['id-ce',
                                                                   17]},
                                'id-ce-subjectDirectoryAttributes': {'type': 'OBJECT '
                                                                             'IDENTIFIER',
                                                                     'value': ['id-ce',
                                                                               9]},
                                'id-ce-subjectKeyIdentifier': {'type': 'OBJECT '
                                                                       'IDENTIFIER',
                                                               'value': ['id-ce',
                                                                         14]},
                                'id-holdinstruction-callissuer': {'type': 'OBJECT '
                                                                          'IDENTIFIER',
                                                                  'value': ['holdInstruction',
                                                                            2]},
                                'id-holdinstruction-none': {'type': 'OBJECT '
                                                                    'IDENTIFIER',
                                                            'value': ['holdInstruction',
                                                                      1]},
                                'id-holdinstruction-reject': {'type': 'OBJECT '
                                                                      'IDENTIFIER',
                                                              'value': ['holdInstruction',
                                                                        3]},
                                'id-kp-OCSPSigning': {'type': 'OBJECT '
                                                              'IDENTIFIER',
                                                      'value': ['id-kp', 9]},
                                'id-kp-clientAuth': {'type': 'OBJECT '
                                                             'IDENTIFIER',
                                                     'value': ['id-kp', 2]},
                                'id-kp-codeSigning': {'type': 'OBJECT '
                                                              'IDENTIFIER',
                                                      'value': ['id-kp', 3]},
                                'id-kp-emailProtection': {'type': 'OBJECT '
                                                                  'IDENTIFIER',
                                                          'value': ['id-kp',
                                                                    4]},
                                'id-kp-serverAuth': {'type': 'OBJECT '
                                                             'IDENTIFIER',
                                                     'value': ['id-kp', 1]},
                                'id-kp-timeStamping': {'type': 'OBJECT '
                                                               'IDENTIFIER',
                                                       'value': ['id-kp', 8]},
                                'id-pe-authorityInfoAccess': {'type': 'OBJECT '
                                                                      'IDENTIFIER',
                                                              'value': ['id-pe',
                                                                        1]},
                                'id-pe-subjectInfoAccess': {'type': 'OBJECT '
                                                                    'IDENTIFIER',
                                                            'value': ['id-pe',
                                                                      11]}}}}


import json
import unittest
from datetime import datetime
from .utils import Asn1ToolsBaseTest
import asn1tools


CODECS = ['ber', 'der', 'jer', 'oer', 'per', 'uper', 'xer']
ALL_CODECS = CODECS + ['gser']


def loadb(encoded):
    return json.loads(encoded.decode('utf-8'))


class Asn1ToolsCodecsConsistencyTest(Asn1ToolsBaseTest):

    maxDiff = None

    def encode_decode_all_codecs(self, type_spec, values):
        spec = (
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= " + type_spec + " "
            + "END"
        )

        foos = []

        for codec in CODECS:
            foos.append(asn1tools.compile_string(spec, codec))

        gser = asn1tools.compile_string(spec, 'gser')

        for value in values:
            decoded = value

            for foo in foos:
                encoded = foo.encode('A', decoded)
                decoded = foo.decode('A', encoded)
                self.assertEqual(type(decoded), type(value))
                self.assertEqual(decoded, value)

            gser.encode('A', decoded)

    def encode_decode_codecs_file(self,
                                  filename,
                                  type_name,
                                  decoded,
                                  encoded):

        for codec, encoded_message in zip(CODECS, encoded):
            foo = asn1tools.compile_files(filename, codec)

            encoded = foo.encode(type_name,
                                 decoded,
                                 check_constraints=True)

            if codec == 'jer':
                self.assertEqual(loadb(encoded), loadb(encoded_message))
            else:
                self.assertEqual(encoded_message, encoded)

            decoded_message = foo.decode(type_name, encoded)
            self.assertEqual(decoded_message, decoded)

        self.assertEqual(foo.decode(type_name, encoded), decoded)

    def encode_decode_codecs(self, type_spec, decoded, encoded):
        spec = (
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= " + type_spec + " "
            + "END"
        )

        for codec, encoded_message in zip(CODECS, encoded):
            foo = asn1tools.compile_string(spec, codec)

            encoded = foo.encode('A',
                                 decoded,
                                 check_constraints=True)

            if codec == 'jer':
                self.assertEqual(loadb(encoded), loadb(encoded_message))
            else:
                self.assertEqual(encoded_message, encoded)

            decoded_message = foo.decode('A', encoded)
            self.assertEqual(decoded_message, decoded)

        self.assertEqual(foo.decode('A', encoded), decoded)

    def test_boolean(self):
        self.encode_decode_all_codecs("BOOLEAN", [True, False])

    def test_integer(self):
        self.encode_decode_all_codecs("INTEGER", [1, 123456789, -2, 0])

    def test_real(self):
        self.encode_decode_all_codecs("REAL", [0.0, 1.0, -1.0])

    def test_null(self):
        self.encode_decode_all_codecs("NULL", [None])

    def test_bit_string(self):
        self.encode_decode_all_codecs("BIT STRING",
                                      [(b'\x58', 5), (b'\x58\x80', 9)])

    def test_octet_string(self):
        self.encode_decode_all_codecs("OCTET STRING", [b'', b'\x12\x34'])

    def test_object_identifier(self):
        self.encode_decode_all_codecs("OBJECT IDENTIFIER", ['1.2.33'])

    def test_enumerated(self):
        self.encode_decode_all_codecs("ENUMERATED { a(0), b(5) }", ['a', 'b'])

    def test_sequence(self):
        self.encode_decode_all_codecs("SEQUENCE { a NULL }", [{'a': None}])

    def test_sequence_of(self):
        self.encode_decode_all_codecs("SEQUENCE OF NULL", [[], [None, None]])

    def test_set(self):
        self.encode_decode_all_codecs("SET { a NULL }", [{'a': None}])

    def test_set_of(self):
        self.encode_decode_all_codecs("SET OF NULL", [[], [None, None]])

    def test_choice(self):
        self.encode_decode_all_codecs("CHOICE { a NULL }", [('a', None)])

    def test_utf8_string(self):
        self.encode_decode_all_codecs("UTF8String", [u'hi'])

    def test_numeric_string(self):
        self.encode_decode_all_codecs("NumericString", [u'123'])

    def test_printable_string(self):
        self.encode_decode_all_codecs("PrintableString", [u'hi'])

    def test_ia5_string(self):
        self.encode_decode_all_codecs("IA5String", [u'hi'])

    def test_visible_string(self):
        self.encode_decode_all_codecs("VisibleString", [u'hi'])

    def test_general_string(self):
        self.encode_decode_all_codecs("GeneralString", [u'hi'])

    def test_bmp_string(self):
        self.encode_decode_all_codecs("BMPString", [u'hi'])

    def test_graphic_string(self):
        self.encode_decode_all_codecs("GraphicString", [u'hi'])

    def test_teletex_string(self):
        self.encode_decode_all_codecs("TeletexString", [u'hi'])

    def test_universal_string(self):
        self.encode_decode_all_codecs("UniversalString", [u'hi'])

    def test_utc_time(self):
        self.encode_decode_all_codecs("UTCTime", [datetime(2020, 3, 12)])

    def test_generalized_time(self):
        self.encode_decode_all_codecs("GeneralizedTime",
                                      [datetime(2021, 3, 12)])

    def test_error_location(self):
        spec = (
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= SEQUENCE { "
            "  a SEQUENCE { "
            "    b CHOICE { "
            "      c SEQUENCE { "
            "        d INTEGER "
            "      } "
            "    } "
            "  } "
            "}"
            "END"
        )

        for codec in ALL_CODECS:
            foo = asn1tools.compile_string(spec, codec)

            with self.assertRaises(asn1tools.EncodeError) as cm:
                foo.encode('A', {'a': {'b': ('c', {})}})

            self.assertEqual(str(cm.exception),
                             "a: b: c: Sequence member 'd' not found in {}.")

    def test_recursive(self):
        spec = (
            "SEQUENCE { "
            "  a B "
            "} "
            "B ::= CHOICE { "
            "  b A, "
            "  c NULL "
            "} "
        )

        self.encode_decode_all_codecs(spec, [{'a': ('b', {'a': ('c', None)})}])

    def test_with_components(self):
        decoded = {
            'a': 1,
            'b': {
                'c': [
                    (
                        'f', True
                    )
                ],
                'd': 2
            }
        }

        encoded = [
            b'\x30\x0d\x80\x01\x01\xa1\x08\xa0\x03\x81\x01\xff\x81\x01\x02',
            b'\x30\x0d\x80\x01\x01\xa1\x08\xa0\x03\x81\x01\xff\x81\x01\x02',
            b'{"a":1,"b":{"c":[{"f":true}],"d":2}}',
            b'\x01\x01\x01\x81\xff\x01\x02',
            b'\x01\x60\x01\x02',
            b'\x01\x60\x20\x40',
            b'<Foo><a>1</a><b><c><f><true /></f></c><d>2</d></b></Foo>'
        ]

        self.encode_decode_codecs_file('tests/files/with_components.asn',
                                       'Foo',
                                       decoded,
                                       encoded)

    def test_integer_min_max(self):
        decoded = 5

        encoded = [
            b'\x02\x01\x05',
            b'\x02\x01\x05',
            b'5',
            b'\x01\x05',
            b'\x01\x05',
            b'\x01\x05',
            b'<A>5</A>'
        ]

        self.encode_decode_codecs('INTEGER (MIN..MAX)',
                                  decoded,
                                  encoded)


if __name__ == '__main__':
    unittest.main()
